#!/usr/bin/python3

import os
import delune
import skitai	

if __name__ == "__main__":
	pref = skitai.pref ()
	pref.use_reloader = 1
	pref.debug = 1
	
	config = pref.config
	config.sched = "0/5 * * * *"	
	config.local = "http://127.0.0.1:5000/v1"
	
	config.remote = os.environ.get ("DELUNE_MIRROR")
	config.enable_mirror = config.remote
	
	config.resource_dir = skitai.joinpath ('resources')
	config.enable_index = True
	
	config.logpath = None
	skitai.trackers ('delune:collection')
	skitai.mount ("/v1", delune, "app", pref)
	skitai.run (	
		workers = 2,
		port = 5000,
		logpath = config.logpath
	)



import os
import threading
import time
import types
import sys
from .. import _delune, memory
from .segment import segmentreaders
from . import query, cache, fetcher
import delune
from delune import binfile, docqueue
from delune.lib import unistr
from delune import filesystem

class Searcher:
	MAINTERN_INTERVAL = 60
	
	def __init__ (self, si, do_init = True):
		self.si = si
		
		# default opt
		self.maxResult = self.si.getopt (max_result = 2000)
		self.numQueryCache = self.si.getopt (num_query_cache = 200)		
		self.estTotal = 0		
		self.numquery = 0
		self.segments = segmentreaders.SegmentReaders (self.si)
		
		self.references = 0
		self.shutdown_level = 0
		self.maintern_time = -1
		self.closed = 1
		self.clean = 0
		self.cond = threading.Condition ()
		self.mutex = threading.Lock ()
		self.cache = cache.Cache (self.numQueryCache)
		self.mod_time = self.si.getModfiedTime ()
		self.queue = None
		self._initiallized = False		
		
		if do_init: 
			self.init ()
	
	def create_queue (self):
		self.queue = docqueue.DocQueue (self.si)		
		
	def init (self):
		self.si.log ("initializing `%s'" % self.si.getAlias ())				
		success = self.do_refresh ()
		if not success:
			self.si.log ("initializing failed `%s'" % self.si.getAlias (), "fail")
		else:	
			self.si.log ("initialized `%s' %s" % (self.si.getAlias (), self.si.getSegmentList ()))
		self._initiallized = True
		return success

	#-----------------------------------------------------------------------------------------
	# maintern functions
	#-----------------------------------------------------------------------------------------
	def ismainternable (self):
		if self.si.lock.islocked ("merge"):
			return 0
		if self.si.lock.isplocked ("duplicate"):
			return 0			
		return 1

	def maintern (self):
		# maintern index
		# another process reading file
		# there is 2 cases when merging and duplicating		
		if time.time () - self.maintern_time > self.MAINTERN_INTERVAL and self.ismainternable ():
			self.maintern_time = time.time ()
			self.queue and self.queue.maintern ()
			if self.need_refresh ():
				self.si.log ("refreshing `%s'" % self.si.getAlias ())				
				self.do_refresh ()
			else:
				#self.si.log ("reloading deldoc `%s'" % self.si.getAlias ())
				self.segments.reload_deleted () # reload del files

	def increference (self):
		self.cond.acquire ()
		while self.shutdown_level == 1: self.cond.wait () #single job working, wait...

		# already shutdowned
		if self.shutdown_level == 2:
			self.cond.release ()
			return 0

		if not self.references:
			try:				
				self.maintern ()
			except MemoryError:
				raise
			except:
				self.si.trace ()

		self.references += 1
		self.cond.release ()
		return 1

	def decreference (self):
		self.cond.acquire ()
		self.references -= 1
		self.numquery += 1
		self.cond.notifyAll ()
		self.cond.release ()
	
	def need_refresh (self):
		mod_time = self.si.getModfiedTime ()		
		if mod_time == -1: # maybe renaming or no segments
			self.clean = 0
			return 0
		
		if self.clean and mod_time == self.mod_time:
			return 0 # not changed
		
		self.si.read () # refreshing new segment infos
		return 1
			
	#-----------------------------------------------------------------------------------------
	# job frameowork
	#-----------------------------------------------------------------------------------------
	def single_job (self, func, levelto, *arg, **karg):
		self.cond.acquire ()
		self.shutdown_level = 1
		while self.references: self.cond.wait () #under using, wait...

		retval = 0
		if self.shutdown_level == 2:
			self.cond.notifyAll ()
			self.cond.release ()
			return 0

		try:
			try: retval = func (*arg, **karg)
			finally: self.shutdown_level = levelto

		except MemoryError:
			self.si.trace ()
			self.cond.notifyAll ()
			self.cond.release ()
			raise

		except:
			self.si.trace ()

		self.cond.notifyAll ()
		self.cond.release ()

		return retval

	def multi_job (self, func, default, *arg, **karg):
		if not self.increference ():
			return [510, 0, 0, []]  #shutdown step or MemroyError

		try:
			try:
				res = func (*arg, **karg)
			finally:
				self.decreference ()

		except MemoryError:
			raise

		except:
			self.si.trace ()
			return default

		else:
			return res

	#-----------------------------------------------------------------------------------------
	# real functions
	#-----------------------------------------------------------------------------------------
	def do_status (self, *arg, **karg):
		segmentinfos = [(x.seg, x.getNumDoc (), x.getDeletedCount (), x.ti.numterm, x.get_delfile_modtime ()) for x in self.segments]
		locks, note = self.si.lock.locks ()
		
		return {
			"version": self.si.version,			
			"lastupdated": max (self.mod_time, segmentinfos and max ([x [-1] for x in segmentinfos]) or 0.),
			"indexdirs": self.si.indexdir,
			"segmentfiles": {"primary": filesystem.get_segment_files (self.si.indexdir)},
			"numquery": self.numquery,
			"N": self.si.getN (),
			"segmentinfos": segmentinfos,
			"segmentsizes": self.si.fs.dirinfo (),
			"locks": locks,
			"note": note,
			# IMP: replicator need current segments modtime
			"segments": (self.si.getSegmentsPath (), self.mod_time)
		}

	def do_remove (self, segs):
		self.segments.remove (segs)
		
	def do_refresh (self, *arg, **karg):
		self.clean = 1		
		if self.si.getModfiedTime () == -1:			
			self.clean = 0
			return 1
			
		try:
			self.segments.load ()

		except:
			# if failed, loading next time
			self.si.trace ()
			self.clean = 0
			return 0
		
		self.cache.clear ()
		self.closed = 0
		self.mod_time = self.si.getModfiedTime ()

		return 1

	def do_close (self, *arg, **karg):
		if self.closed: return 1		
		self.segments.close ()
		self.si.close ()
		self.si = None
		self.closed = 1
		return 1
	
	def do_commit (self):
		try:			
			self.segments.commit ()
			self.cache.clear ()
		finally:
			self.si.fs.dirty ()
			
	def do_delete (self, qs, lang = "en", analyze = 1, commit = True):
		try:
			try:
				plock = self.si.plock
				if plock is not None: plock.acquire ()  # process lock			
				result = self.do_query (qs, 0, 2000000000, "", 0, lang = lang, analyze = analyze, data = 0, limit = 0)
				
				if "result" in result:
					for record in result ["result"]:
						seg, docid, extra, score = record
						self.segments [seg].delete (int (docid))						
					
					if commit:
						self.do_commit ()
						
			finally:
				if plock is not None: 
					plock.release ()

		except:
			self.si.trace ()
			#return [500, 0, 0, ""]
			return {"code":500, "err": "Deletion Error"}
		
		#return [200, 0, result [2], result [3]]
		return {"code":200, "total": result ["total"], "time": result ["time"]}		
		
	def do_query (self, qs, offset = 0, fetch = 10, sort = "", summary = 30, returns = "", nthdoc = 0, lang = "un", analyze = 1, data = 1, limit = 1, *arg, **karg):
		try:
			offset = int (offset)
			fetch = int (fetch)
			summary = int (summary)
			analyze = int (analyze)
			data = int (data)
			limit = int (limit)
			nthdoc = int (nthdoc)
			qs = unistr.makes (qs)
			
		except:
			#return [505, 0, 0, ""]
			return {"code": 501, "err": "Arguments Error", "total": 0}
		
		if not qs.strip ():
			return {"code": 502, "err": "No Keyword", "total": 0}
			#return [210, 0, 0, ""]
		if limit and self.maxResult and offset + fetch > self.maxResult: 
			return {"code": 503, "err": "Exceed Limit", "total": 0}

		#print ("#############", qs)
		#if isinstance (qs, str): 
		#	qs = qs.encode ("utf8")

		cachekey = (qs, lang, sort)
		cached = self.cache.get (cachekey)
		
		result = {"code": 200, "time": 0, "total": 0}
		if cached and (len (cached [-1]) >= offset + fetch or len (cached [-1]) == cached [0]):
			totalcount, sortindex, sortorder, highlights, bucket = cached
			result ["msg"] = "Hit Cache"
			result ["total"] = totalcount
			result ["regex"] = query.makeHighLightRE (highlights)
			result ["sorted"] = [sortindex, sortorder]
			result ["time"] = 0.0
			if data:
				result ["result"] = self.do_fetch (bucket  [offset:offset + fetch], summary, highlights, returns, nthdoc)								
			else:
				result ["result"] = bucket  [offset:offset + fetch]
			return result			
		
		q = query.Query (self.segments, qs, lang, offset, fetch, sort, analyze and self.si.analyzer or None, self.estTotal)
		try:
			q.query ()			
			result ["total"] = q.totalcount
			result ["regex"] = query.makeHighLightRE (q.highlights)
			result ["time"] = q.duration
			result ["sorted"] = [q.sortindex, q.sortorder]
									
			if data:
				result ["result"] = self.do_fetch (q.bucket  [offset:offset + fetch], summary, list(q.highlights.keys ()), returns, nthdoc)
				#result = [200, q.duration, q.totalcount, query.makeHighLightRE (q.highlights)] + fetched
			else:
				result ["result"] = q.bucket  [offset:offset + fetch]				
				#result = [200, [q.sortindex, q.sortorder], q.totalcount, query.makeHighLightRE (q.highlights)] + fetched

			if offset == 0:
				self.cache.set (cachekey, (q.totalcount, q.sortindex, q.sortorder, list(q.highlights.keys ()), q.bucket [:100]))

			if q.esegments:
				self.do_remove (q.esegments)

		finally:
			q.close ()
			del q

		return result

	def do_fetch (self, bucket, summary, highlights, returns, nthdoc, *arg, **karg):
		f = fetcher.Fetcher (self.segments, bucket, summary, highlights, returns, nthdoc)		
		try:
			result = f.fetchall ()
		finally:			
			f.searcher = None			
			f.close ()		
		return result

	#-----------------------------------------------------------------------------------------
	# interface functions
	#-----------------------------------------------------------------------------------------
	def close (self, *arg, **karg):
		return self.single_job (self.do_close, 2, *arg, **karg)
		
	def refresh (self, *arg, **karg):
		self.si.read () # force refresh
		return self.single_job (self.do_refresh, 0, *arg, **karg)
	
	def delete (self, *arg, **karg):
		return self.single_job (self.do_delete, 0, *arg, **karg)
	
	def commit (self, *arg, **karg):
		return self.single_job (self.do_commit, 0, *arg, **karg)
	
	def status (self, *arg, **karg):
		default = {}
		return self.multi_job (self.do_status, default, *arg, **karg)

	def query (self, *arg, **karg):		
		default = {"code": 500, "err": "Default Error", "total": 0}
		return self.multi_job (self.do_query, default, *arg, **karg)

	def fetch (self, *arg, **karg):
		default = []
		return self.multi_job (self.do_fetch, default, *arg, **karg)
	

if __name__ == "__main__":
	from delune import indexinfo
	from delune.lib import confparse, logger
	#from delune.searcher import memory
	import memory
	
	memory.create (1, 32768, 1024000, "segment", logger.screen_logger ())
	inf = indexinfo.IndexInfo (confparse.ConfParse (r"d:\bladepub\proto\etc\col\mailservice"))
	s = Searcher (inf)
	try:
		s.init ()
		print(s.query ("computer", 0, 0))

	finally:
		s.close ()
		




import os
import sys
import types
from delune import _delune
from delune.searcher.segment import segmentreader
from delune.searcher import collection
from delune.util import util
from . import segment
import pickle as pk


class FeatureClassInfos:
	def __init__ (self, path):
		self.path = path
		self.infos = {}
	
	def __getitem__ (self, k):
		return self.infos [k]
	
	def __setitem__ (self, k, v):
		self.infos [k] = v
	
	def truncate (self):		
		self.infos = {}	
		self.path = None	
	
	def load (self):		
		if not self.path: 
			return
		if not os.path.isfile (self.path):
			return
		f = open (self.path, "rb")
		self.infos = pk.load (f)
		f.close ()
		
	def save (self, path = None):
		if path: self.path = path
		f = open (self.path, "wb")
		pk.dump (self.infos, f, 1)
		f.close ()
		
	def close (self):
		del self.infos
		self.infos = {}


class ComposedSegmentWriter:
	def __init__ (self, si):
		self.si = si		
		self.logger = self.si.logger
		
		self.featureClassInfoCache = FeatureClassInfos (self.getSegmentBasePath () + "cfi")
		self.DFCache = {}
		self.mpool = _delune.MemoryPool (1, 32768, 50)
		self.mem = self.mpool.get (0)
		
		self.ts_si = collection.Collection (os.path.join (self.si.fs.getmaster (), "%s.ts" % self.si.alias), 'w', self.si.logger, version = self.si.version)
		self.ts_reader = segmentreader.SegmentReader (self.ts_si, self.ts_si.getSegmentList ()[0])
		
		self.fdnoclass = self.ts_si.getFdnoByName ("label")
		self.fdnoterm = self.ts_si.getFdnoByName ("default")
		
		self.basesegmentpath = None	
		self.dirty = 0
	
	def getSegmentBasePath (self):
		try:
			seg = self.si.fs.segments ()[-1]
		except IndexError:
			return os.path.join (self.si.fs.getmaster (), "0.")
		return os.path.join (self.si.fs.getmaster (), str (seg) + ".")
				
	def close (self):
		if self.mpool:
			self.mpool.close ()
			self.mpool = None
		if self.ts_reader:
			self.ts_reader.close ()
			
	def getIndexedNumDoc (self):
		return self.ts_reader.si.getN ()
			
	def getSegmentBulkReader (self):
		return segmentreader.SegmentBulkReader (self.ts_si, self.ts_reader.si.getSegmentList ()[0])
		
	def getDocument (self, docid):
		fd, summary = self.ts_reader.getDocument (self.mem, docid)
		return pk.loads (fd) [:2]
	
	def getDF (self, term, fdno):
		return self.getTermInfo (term, fdno) [0]
		
	def getTermInfo (self, term, fdno):
		try:
			return self.DFCache ["%s:%s" % (fdno, term)]
		except KeyError:
			value = self.ts_reader.getTermInfo (self.mem, term, fdno)
			self.DFCache ["%s:%s" % (fdno, term)] = value
			return value
		
	def getTermClassInfo (self, term, pools):
		hits = _delune.Compute (self.mem)
		hits.newscan ()
		hits.setfreq (1)
		
		df, doff, poff, skip, plen = self.ts_reader.getTermInfo (self.mem, term, self.fdnoterm)		
		self.ts_reader.readPosting (self.mem, df, doff, poff, skip, plen, -1, -1, 0)
		hits.set ()
		
		cfi = {}
		tdf = 0
		numpool = 0
		for pool in pools:
			df, doff, poff, skip, plen = self.ts_reader.getTermInfo (self.mem, pool.name, self.fdnoclass)			
			self.ts_reader.readPosting (self.mem, df, doff, poff, skip, plen, -1, -1, 0)
			df = hits.intersect ()
			tf = 0
			if df:
				tdf += df
				r = hits.hitdoc (0)
				for seg, docid, freq, score in r:
					tf += freq	
				cfi [numpool] = (df, tf)
			hits.reuse ()
			numpool += 1						
		hits.close ()
		return tdf, cfi
	
	def getSegmentPath (self):
		return self.basesegmentpath		
	
	def featureClassInfo  (self, term, corpus): 
		try:
			cfi = self.featureClassInfoCache [term]
			fi = [(0,0)] * corpus.numpools
			tdf = 0
			for index, info in list(cfi.items ()):
				tdf += info [0]
				fi [index] = info
			return tdf, fi			
		except KeyError:
			pass
		tdf, cfi = self.getTermClassInfo (term, corpus.pools)
		self.featureClassInfoCache [term] = cfi
		return self.featureClassInfo (term, corpus)
	
	def flush (self):			
		if self.dirty:
			self.si.log ("flushing train data...")
			self.si.flush (new = 1)			
		else:
			self.si.flush (new = 0)
	
	def writeTermTermInfo (self, trainsegment, terms):
		numterm = len (terms)
		self.si.log ("saving %d terms co-occurences..." % (numterm,))
		hits = _delune.Compute (self.mem)
		cnt = 0
		poslist = []
		for term in terms:
			pos = trainsegment.co.tell ()
			if cnt % 100 == 0:
				self.si.log ("writing %d/%d terms co-occurences..." % (cnt, numterm))
			cnt += 1
			hits.newscan ()
			colist = []
			df, doff, poff, skip, plen = self.ts_reader.getTermInfo (self.mem, term, self.fdnoterm)		
			self.ts_reader.readPosting (self.mem, df, doff, poff, skip, plen, -1, -1, 0)
			hits.set ()
			rindex = cnt
			for term2 in terms [cnt:]:
				df, doff, poff, skip, plen = self.ts_reader.getTermInfo (self.mem, term2, self.fdnoterm)			
				self.ts_reader.readPosting (self.mem, df, doff, poff, skip, plen, -1, -1, 0)
				df = hits.intersect ()
				if df:
					colist.append ((rindex, df))
				rindex += 1
				hits.reuse ()
			df, endpos = trainsegment.co.write (colist, 1)
			poslist.extend ([pos, df])
		hits.close ()		
		trainsegment.ci.writelist (poslist, 4)
			
	def commit (self, corpus):
		if not self.dirty:
			return
		
		self.si.createSegments ()	# always rebuild
		self.si.segments.pools = [ pool.name for pool in corpus.pools ]
		self.si.segments.numpool = corpus.numpools
		self.si.segments.features = {}
		self.si.segments.numvoca = len (corpus.terms)
		self.si.segments.N = corpus.N

		self.si.log ("creating new segment...")
		newseg = self.si.getNewSegment  ()
		
		trainsegment = segment.Segment (self.si.fs.new (newseg), newseg, "w", version = self.si.version)
		trainsegment.open ()
		
		frqposition = trainsegment.tf.tell ()
		df, skip = trainsegment.tf.write ([(pool.df, pool.tf) for pool in corpus.pools], 0, 1)
		trainsegment.ti.add ("corpus", 0, df, frqposition, 0, skip - frqposition, 0)

		self.si.log ("writing terms...")
		c = 0
		t = len (corpus.terms)
		terms = list(corpus.terms.keys ())
		terms.sort ()
		for term in terms:
			if c % 1000 == 0:
				self.si.log ("writing terms %d / %d" % (c, t))
			frqposition = trainsegment.tf.tell ()
			tdf, fi = self.featureClassInfo (term, corpus)
			if not tdf: continue
			df, skip = trainsegment.tf.write (fi)
			if df:
				trainsegment.ti.add (term, 1, df, frqposition, 0, skip - frqposition, 0)
			c += 1
		
		self.basesegmentpath = trainsegment.get_base_path ()
		self.si.log ("writing feature-class information...")
		self.featureClassInfoCache.save (self.basesegmentpath + "cfi")
		self.writeTermTermInfo (trainsegment, terms)
		trainsegment.close ()
		
		self.si.addSegment (newseg, corpus.N)
		self.si.write ("segments.new")
		
		self.si.log ("saving temporary corpus")
		f = open (self.basesegmentpath + "cps", "wb")
		pk.dump (corpus, f, 1)
		f.close ()		
		self.si.log ("writing, done.")


class ComposedSegmentFeaturesetWriter (ComposedSegmentWriter):
	def __init__ (self, si):
		self.si = si		
		self.logger = self.si.logger
		
		self.featureClassInfoCache = FeatureClassInfos (self.getSegmentBasePath () + "cfi")
		self.featureClassInfoCache.load ()
		self.dirty = 1
	
	def get_corpus (self):
		path = os.path.join (self.getSegmentBasePath () + "cps")
		with open (path, "rb") as f:
			corpus = pk.load (f)
		return corpus	
		
	def close (self):
		pass

	


from delune import _delune
from .util import htmlstrip
from . import abstractAnalyzer
import re

RX_DIGIT = re.compile ("(\s|^)[^a-zA-Z]+(\s|$)")
	
class Analyzer (abstractAnalyzer.Analyzer):
	def __init__ (self, max_term = 8, numthread = 1, **karg):
		self.hstriper = htmlstrip.build_parser ()
		self._strip_html = False # for fast detect
		self._make_lower_case = False
		self._contains_alpha_only = False
		self.karg = karg
		self.options = {
			"biword": 0,
			"ngram": 1,
			"stem_level": 1,
			"stopwords": [],
			"endwords": [],
			"stopwords_case_sensitive": 1,
			"ngram_no_space": 0,	
			"strip_html": 0,
			"make_lower_case": 0,
			"contains_alpha_only": 0	
		}
		abstractAnalyzer.Analyzer.__init__ (self, max_term, numthread)
		self.setopt (**karg)
	
	def createTSAnalyzers (self, num):
		for i in range (num):
			# stem-level=weak, ngram=bi, default stopwords list			
			f = _delune.Analyzer (self.max_term, self.karg.get ("stem_level", 1), self.karg.get ("ngram", 2), self.karg.get ("biword", 0))
			f.set_stopwords ()
			self.ats.append (f)
	
	#------------------------------------------------
	# set options 
	#------------------------------------------------
	def getopt (self, **karg):
		for k, v in list(karg.items ()):	
			return self.options.get (k, v)
						
	def setopt (self, **karg):
		for k, v in list(karg.items ()):
			self.apply_option (k, v)
			
	def apply_option (self, name, value):
		try:
			setfunc = getattr (self, "set_" + name)
		except AttributeError:
			raise NameError("Unknown Option `%s`" % name)
		
		self.options [name] = value
		setfunc (value)
			
	#------------------------------------------------
	# exported methods
	#------------------------------------------------			
	def strip_html (self, document):
		try:
			return self.hstriper.parse (document)
		except UnicodeDecodeError:
			return self.hstriper.parse (self.to_ascii (document))
		except UnicodeEncodeError:
			return self.hstriper.parse (self.remove_surrogate_escaping (document))	
			
	def preprocess (self, document):
		if self._strip_html:
			document = self.strip_html (document)		
		if self._make_lower_case:
			document = document.lower ()	
		if self._contains_alpha_only:
			document = RX_DIGIT.sub (" ", document)
		return document
		
	def to_ascii (self, document):
		return document.encode ("ascii", "ignore").decode ("utf8")
		
	def remove_surrogate_escaping (self, document):
		return document.encode ("utf8", "ignore").decode ("utf8")
		
	def formalize (self, document):
		try:
			return self.ats [self.get_tid ()].formalize (document)
		except UnicodeDecodeError:	
			return self.ats [self.get_tid ()].formalize (self.to_ascii (document))
		except UnicodeEncodeError:	
			return self.ats [self.get_tid ()].formalize (self.remove_surrogate_escaping (document))
				
	def index	(self, document, lang = "en"):
		document = self.preprocess (document)
		return self.ats [self.get_tid ()].analyze (document, lang)
	
	def freq (self, document, lang = "en"):		
		k = self.index (document, lang)
		if k:
			return [(x [0], len (x [1])) for x in list(k.items ())]		
		return []	
	term = freq
	
	def tokenize (self, document, lang = "en"):
		# no preprocess
		try:
			return self.ats [self.get_tid ()].stem (document, lang)
		except UnicodeDecodeError:	
			return self.ats [self.get_tid ()].stem (self.to_ascii (document), lang)
		except UnicodeEncodeError:	
			return self.ats [self.get_tid ()].stem (self.remove_surrogate_escaping (document))	
	query = tokenize
	
	def stem (self, document, lang = "en"):
		document = self.preprocess (document)
		return self.tokenize (document)
	
	def count_stopwords (self, document, lang = "en"):
		document = self.preprocess (document)
		try:
			return self.ats [self.get_tid ()].count_stopwords (document, lang)
		except UnicodeDecodeError:	
			return self.ats [self.get_tid ()].count_stopwords (self.to_ascii (document), lang)
		except UnicodeEncodeError:	
			return self.ats [self.get_tid ()].formalize (self.remove_surrogate_escaping (document))	
		
	#------------------------------------------------
	# handle options 
	#------------------------------------------------	
	def set_contains_alpha_only (self, flag):
		self._contains_alpha_only = flag
		
	def set_strip_html (self, flag):
		self._strip_html = flag
	
	def set_make_lower_case (self, flag):
		self._make_lower_case = flag
		
	def set_stopwords (self, wordlist):
		if not wordlist: return
		for an in self.ats:
			an.set_stopwords (wordlist)
			
	def set_endwords (self, wordlist):
		if not wordlist: return
		for an in self.ats:
			an.set_endwords (wordlist)		
	
	def set_stem_level (self, stem_level):
		for an in self.ats:
			an.set_stem_level (stem_level)
		
	def set_ngram_no_space (self, flag):
		for an in self.ats:
			an.set_ngram_ignore_space (flag)	
	
	def set_stopwords_case_sensitive (self, flag):
		for an in self.ats:
			an.set_stopwords_case_sensitive (flag)	
				
	def set_ngram (self, ngram):
		for an in self.ats:
			an.set_ngram (ngram)	
	
	def set_biword (self, flag):
		for an in self.ats:
			an.set_biword (flag)
		


from delune import binfile
from delune.lib import pathtool
import os
import shutil
import random
import threading
import glob
import time

class DocQueue:
	quename = "que"
	def __init__ (self, si):
		self.si = si
		self.que_dir = os.path.join (self.si.fs.getmaster (), '.que')
		pathtool.mkdir (self.que_dir)
		self._tlock = threading.RLock ()
		self._plock = self.si.plock # process lock
		self.writer = None
		self.current = None
		self.commit_all ()
		self.last_written = time.time ()
		
	def maintern (self):
		with self._tlock:
			commit = time.time () > self.last_written + 300
		if commit:
			self.commit ()
			
	def __detect_max_seq (self):	
		while 1:
			rn = random.randrange (100000, 999999)
			if not os.path.isfile (os.path.join (self.que_dir, "%s.%d.lock" % (self.quename, rn))) and not os.path.isfile (os.path.join (self.que_dir, "%s.%d" % (self.quename, rn))):
				return rn
	
	def __commit (self, temp, remove = False):
		if remove or os.path.getsize (temp) == 0:
			os.remove (temp)
		else:	
			shutil.move (temp, temp [:-5])

	def commit_all (self):		
		for uncommit in os.listdir (self.que_dir):
			if uncommit.startswith (self.quename + ".") and uncommit.endswith (".lock"):
				self.__commit (os.path.join (self.que_dir, uncommit))
		
	def __call__ (self, method, doc):
		self.add (method, doc)
	
	def open (self):	
		self.current = self.__detect_max_seq ()
		self.writer = binfile.BinFile (os.path.join (self.que_dir, "%s.%d.lock" % (self.quename, self.current)), "w")
	
	def truncate (self):
		self.close ()
		with open (os.path.join (self.que_dir, "%s.truncate" % self.quename), "w") as f:
			f.write ("DO NOT DELETE")
		
	def close (self, remove = 0):
		if not self.writer: 
			return
		os.fsync (self.writer.fileno ())	
		self.writer.close ()
		self.writer = None
		self.__commit (
			os.path.join (self.que_dir, "%s.%d.lock" % (self.quename, self.current)), 
			remove
		)
			
	def rollback (self):
		with self._tlock:
			self.close (1)
			self.open ()
	
	def commit (self):
		with self._tlock:
			self.close ()
			
	def add (self, method, doc):
		with self._tlock:
			self.last_written = time.time ()
			if self.writer is None:			
				self.open ()
			self.writer.writeVInt (method)
			self.writer.writeZBytes (doc)



import skitai
import sys, os
from .package import mirror, indexer

def bootstrap (pref):	
	config = pref.config
	
	script = None
	logpath = config.get ("logpath")
	logopt = logpath and "--log=%s" % logpath or ""
	if config.get ("enable_mirror", False):
		script = "%s %s %s" % (mirror.__file__, logopt, config.remote)
	elif config.get ("enable_index", False):
		script = "%s %s %s" % (indexer.__file__, logopt, config.resource_dir)
	
	if script:		
		skitai.cron (config.sched, r"%s %s%s" % (sys.executable, script, ''))



from . import segment
from . import segmentreader
from ... import _delune
import time

class SegmentMerger (segment.Segment):
	def __init__ (self, si, truncate = False):
		self.si = si
		self.truncate = truncate
		self.delay_sec, self.delay_factor = self.si.getopt (lazy_merge = (0, 3))
		if not self.delay_factor:
			self.delay_factor = 3
		newseg = self.si.getNewSegment ()
		segment.Segment.__init__ (self, si.fs.new (newseg), newseg, 'w', version = self.si.version)
		self.mergeinfo = _delune.MergeInfo ()
		self.smis = self.mergeinfo.get ()
		self.segments = []
		self.deletables = []
	
	def close (self):
		if self.mergeinfo:
			self.mergeinfo.close ()
			self.mergeinfo = None
		segment.Segment.close (self)
			
	def addSegment (self, seg):
		reader = segmentreader.SegmentBulkReader (self.si, seg)
		self.deletables.append (reader)
		if reader.numDoc and not self.truncate:
			self.segments.append (reader)
		
	def merge (self):
		self.segments.sort (key = lambda x: int (x.seg))
		
		for segment in self.segments:
			self.mergeinfo.add (*segment.getMergeInfo ())
			
		self.tf.setsmis (self.smis)	
		self.fd.setsmis (self.smis)
		self.sm.setsmis (self.smis)
		
		self.mergeDocument ()		
		self.mergeSortMap ()		
		self.mergeTermInfo ()
		self.flush ()
		
	def flush (self):
		for reader in self.deletables:
			seg = reader.seg
			reader.close ()
			self.si.removeSegment (seg)
						
		self.si.addSegment (self.seg, self.numDoc)
		self.close ()
		self.si.flush ()
		
	def mergeTermInfo (self):
		delay_factor = self.delay_factor * 10000
		slots = [None] * len (self.segments)
		i = 0
		for segment in self.segments:
			slots [i] = segment.advanceTermInfo ()
			i += 1			
		
		#slots = slots [:1]
		termnum = 0
		while 1:
			fslots = [_f for _f in slots if _f]
			if not fslots: break
			#slots = slots [:1]
			
			miti = min (fslots)			
			
			_term, _fdno = miti.term, miti.fdno			
			i = 0
			last_docid = 0
			tdf = 0
			frqposition, prxposition = self.tf.tell ()
			for ti in slots:
				if ti and miti == ti:
					tdf, skip, pskip, last_docid = self.tf.merge (i, ti.df, ti.frqPosition, ti.prxPosition, ti.prxLength, tdf, last_docid)
					try: 
						slots [i] = self.segments [i].advanceTermInfo ()					
					except IndexError: 
						slots [i] = None
				i += 1
			
			# only when has docments
			if not tdf:
				continue
				
			self.ti.add (_term, _fdno, tdf, frqposition, prxposition, skip - frqposition, pskip - prxposition) #skipdelta: skip - position
			self.tf.commit ()			
			termnum += 1
			if termnum and termnum % delay_factor == 0:
				time.sleep (self.delay_sec)
			
	def mergeDocument (self):
		delay_factor = self.delay_factor * 1000
		for seg in range (len (self.segments)):
			for docid in range (self.segments [seg].numDoc):
				self.numDoc += self.fd.merge (seg, docid)
				if self.numDoc and self.numDoc % delay_factor == 0:
					time.sleep (self.delay_sec)
			
	def mergeSortMap (self):
		for fdno, _size in self.si.getSortMapFields ():
			for seg in range (len (self.segments)):
				pointer, size = self.segments [seg].getSortMapPointer (fdno)			
				self.sm.merge (seg, fdno, pointer, size)			
			time.sleep (self.delay_sec)


import delune
from pprint import pprint
import json
import codecs
import os
import requests

class Collection:
	def __init__ (self, ratio = 0.9, front = 1):		
		self.ratio = ratio
		self.front = front
		if ratio == 1.0 and not front:
			self.ratio = 0.9			
		self.skipped = 0
		self.skipping = 1
		self.__totaldoc = 0
		self.__numdoc = 0
		self.setup ()
	
	def __skip (self):	
		self.skipping = 1	
		if self.offset:
			if self.offset > self.__totaldoc:				
				raise IndexError ("offset is larger than total documents")
			while self.__numdoc < self.offset:
				self.get ()
				self.__numdoc += 1
		self.skipping = 0
		
	def __iter__ (self):
		return self
	
	def __next__ (self):
		if not self.skipped:
			self.__skip ()
			self.skipped = 1
			
		doc = self.get ()
		self.__numdoc += 1
			
		if doc is None or self.__numdoc >= self.__totaldoc or self.__numdoc >= (self.offset + self.limit):
			self.teardown ()
			raise StopIteration			
		return doc
	
	def reget (self):
		self.__numdoc += 1
		return self.get ()
		
	def get_total (self):
		return self.__totaldoc
	
	@property
	def index (self):
		return self.__numdoc
		
	def set_total (self, num):
		self.__totaldoc = num
		if self.front:
			self.offset = 0
			self.limit = int (self.__totaldoc * self.ratio)
		else:
			self.offset = int (self.__totaldoc * self.ratio)
			self.limit = self.__totaldoc - self.offset		
	
	def setup (self):
		pass
			
	def teardown (self):
		pass
	
	def get (self):
		raise NotImplementedError	
		
		
class GuessAndCheck:
	def __init__ (self, colopt, col_class, trainset_ratio, trialset, resource_dir = ""):
		self.colopt = colopt	
		self.col_class = col_class
		self.trainset_ratio = trainset_ratio		
		self.params = trialset
		self.resource_dir = resource_dir
		self.logger = None
	
	def set_logger (self, logger):
		self.logger = logger		
		
	def get_model (self, mode):	
		mdl = delune.model (
			[os.path.join (self.resource_dir, d) for d in self.colopt ['data_dir']], 
			mode, 
			delune.standard_analyzer (max_term = 10000, **self.colopt ['analyzer'])
		)
		return mdl
		
	def get_learner (self, mode = delune.MODIFY):			
		return self.get_model (mode).get_learner ()
	
	def get_classifier (self):			
		return self.get_model(delune.READ).get_classifier ()
	
	def add_document (self, learner, docs):
		if type (docs) is not list:
			docs = [docs]
		for doc in docs:				
			learner.add_document (doc)
				
	def index (self):
		try: os.remove (self.get_file ())
		except OSError: pass
		learner = self.get_learner (delune.CREATE)
		for docs in self.col_class (self.trainset_ratio, 1):			
			self.add_document (learner, docs)
		learner.close ()
	
	def append (self):
		learner = self.get_learner (delune.APPEND)
		for docs in self.col_class (self.trainset_ratio, 0):
			self.add_document (learner, docs)
		learner.close ()
		
	def optimize (self):
		learner = self.get_learner ()
		learner.optimize ()
		learner.close ()
			
	def listbydf (self, dfmin = 0, dfmax = 0):
		learner = self.get_learner ()			
		learner.listbydf (dfmin, dfmax)
		learner.close ()
	
	def guess (self, document, lang = "en"):
		classifier = self.get_classifier ()
		classifier.setopt (delune.ROCCHINO, topdoc = 200)
		rs = []
		for cl in (delune.META, delune.NAIVEBAYES, delune.FEATUREVOTE, delune.ROCCHINO, delune.MULTIPATH, delune.TFIDF, delune.SIMILARITY):
			rs.append (classifier.guess (document, lang = "en", cl = cl))
		pprint (rs)		
		classifier.close ()
	
	def build (self, dfmin = 0, dfmax = 0):
		learner = self.get_learner ()	
		r = learner.build (dfmin, dfmax)
		learner.close ()
		return r
	
	def train (self, selector = delune.IG, select = 2000):
		learner = self.get_learner ()	
		r = learner.train (selector = selector, select = select, orderby = delune.MAX, dfmin = 0, dfmax = 0)
		learner.close ()			
		return r
		
	def check (self, cl = delune.NAIVEBAYES):
		classifier = self.get_classifier ()
		classifier.setopt (delune.ROCCHINO, topdoc = 200)
				
		total = 0
		matched = 0		
		unclaissified = 0
		
		for docs in self.col_class (self.trainset_ratio, 0):
			if type (docs) is not list:
				docs = [docs]
			
			for doc in docs:	
				total += 1
				rs = classifier.guess (doc.text, cl = cl, lang = doc.lang)
				if not rs ["result"]:
					unclaissified += 1				
					continue
				
				if len (rs ["result"]) and rs ["result"][0][0] == doc.label:	
					matched += 1	
					if self.logger and matched % 100 == 0:
						self.logger ("%2.1f%% (%d droped / %d documents)" % (matched / (total - unclaissified) * 100, unclaissified, total), 'info')
			
		classifier.close ()
		return (matched / (total - unclaissified), unclaissified, total)		
	
	def test (self, cl = delune.NAIVEBAYES):
		result = self.check (cl)
		print ("accuracy: %1.4f, unclassified: %d, total documents: %d" % tuple (result))
		
	def get_file (self):
		data_dir = self.colopt ['data_dir']
		if type (data_dir) is list:
			data_dir = data_dir [0]
		return os.path.join (self.resource_dir, data_dir, 'prof.gac')
			
	def save (self, obj):
		with codecs.open (self.get_file (), "w")	as f:
			f.write (json.dumps (obj))
	
	def load (self):		
		fn = self.get_file ()
		if not os.path.isfile (fn):
			return {}
		with codecs.open (fn)	as f:
			return json.loads (f.read ())
	
	def commit (self, index):
		prev = self.load ()
		params = prev.get ('params')
		if not params:
			raise SystemError ('You should run trials first')
		if 'optimal' in prev:
			raise SystemError ('Already committed')
		try:
			dfmin, dfmax, selector, select, cl = params [index]
		except IndexError:
			raise SystemError ('Invalid parameter index')
				
		self.build (dfmin, dfmax)
		self.train (selector, select)
		self.optimize ()
		prev ['optimal'] = params [index]
		self.save (prev)
		
	def runset (self):
		current_dfs = (-1, -1)
		prev = self.load ()
		if prev.get ('current_dfs'):
			current_dfs = tuple (prev.get ('current_dfs'))
			
		results = []
		trial = 0
		total, pruned = 0, 0		
		for dfmin, dfmax, selector, select, cl in self.params:			
			if current_dfs != (dfmin, dfmax):
				total, pruned = self.build (dfmin, dfmax)
				current_dfs = (dfmin, dfmax)				
			features = self.train (selector, select)
			results.append ((trial, self.check (cl), (features, total, total - pruned, dfmin, dfmax, cl, selector, select)))
			trial += 1
			
		self.save ({
			'results': results,
			'current_dfs': list (current_dfs),
			'params': self.params
		})
		self.status ()
	
	def status (self):
		res = self.load ()		
		print ('------------------')
		print ('testing result')
		print ('------------------')
		for trial, result, options in res ["results"]:			
			print ("#%2d" % trial, end = " ")
			print ("accuracy: %1.4f, unclassified: %d, total documents: %d" % tuple (result))			
			print (
				"    - features: %d, terms: %d -> %d, dfmin: %d, dfmax: %d\n" 
				"    - classifier: %s, selector: %s, select: %s"
				% tuple (options)
			)
			print ()

def usage ():
	import sys
	
	print ("""
Usage
-----

	%s command [parameters]	


Commands and Required Parameters
--------------------------------
	
	status
	index
	runset
	append
	commit trial_index    ex. commit 9
	
	test [classifier=naivebayes]: test accuracy by current options
	  - classifiers: naivebayes, featurevote, tfidf, similarity, rocchino, multipath, metaclassifier
	list min_df max_df    ex. list 5 0
	guess text_to_classify    ex. guess "computer service"
	register alias


Formal Sequence
---------------

	1. Make class for poducing documents
	2. Run index
	3. Run list 0 0
	4. Make trials parameters
	5. Run runset
	6. Run append
	7. Run commit [optimal parameter index]
	8. Run register alias
	
	""" % sys.argv [0])
	
	sys.exit ()
	
	
def handle_command (gac, log = None):
	import getopt, sys
	from delune.lib import logger
	
	argopt = getopt.getopt(sys.argv[1:], "", [])
	log = log or logger.screen_logger ()	
	delune.configure (1, log)	
	gac.set_logger (log)	
	
	params, cmd = argopt
	if not cmd:
		usage ()
				
	params, cmd = tuple (cmd [1:]), cmd [0]	
	
	if cmd == "index":
		gac.index ()
	
	elif cmd == "guess":
		gac.guess (" ".join (params))
		
	elif cmd == "status":
		gac.status ()
	
	elif cmd == "runset":
		gac.runset ()
	
	elif cmd == "list":
		try:
			gac.listbydf (int (params [0]), int (params [1]))
		except IndexError:
			usage ()	
	
	elif cmd == "append":	
		gac.append ()		
	
	elif cmd == "test":	
		if params:
			cl = params [0]
		else:
			cl = delune.NAIVEBAYES
		gac.test (cl)
		
	elif cmd == "commit":	
		try:
			gac.commit (int (params [0]))
		except IndexError:
			usage ()		
	
	elif cmd == "register":
		try:
			url = params [0]
			r = requests.post (url, data = json.dumps (gac.colopt))
			print ("register: %s" % url)
			print ("------------------------------")
			if r.text:
				print ("error: ", r.text)
			else:	
				print ("success: move/copy model into app resource dir")			
		except IndexError:
			usage ()
			
	delune.shutdown ()




import delune
from delune import _delune
import threading

class Analyzer:
	def __init__ (self, max_term = 8, numthread = 1):
		self.max_term = max_term
		self.numthread = numthread
		self.ats = []
		self.createTSAnalyzers (numthread)
		self.closed = False
		
	def __len__ (self):
		return len (self.ats)
		
	def get_tid (self):
		try: return threading.currentThread ().getId ()
		except AttributeError: return -1
	
	def createTSAnalyzers (self, num):
		pass
	
	def close (self):
		if self.closed: return
		for each in self.ats:
			each.close ()
		self.closed = True
		self.ats = []

	def generalize (self, t):
		return t.replace ("\x00", "")
	
	def transform (self, document):
		return document

	def analyze (self, document, lang, fdtype):		
		if fdtype == delune.TEXT:
			return self.index (document, lang)
		elif fdtype == delune.TERM:
			return self.freq (document, lang)
		else:
			return self.tokenize (document, lang)

	def index	(self, document, lang = "en"):
		raise NotImplementedError

	def freq (self, document, lang = "en"):
		raise NotImplementedError

	def tokenize (self, document, lang = "en"):
		raise NotImplementedError



PLUS = '+'
MINUS = '-'
TIMES = '*'
DIVIDE = '/'
LPAREN = '('
RPAREN = ')'
EOF = 'EOF'
SOP = {"and": TIMES, "not": MINUS, "or": PLUS}

class QueryLexer:
	def __init__ (self, stream, allow_or = 1):
		self.stream = stream.strip ().lower ()
		self.allow_or = allow_or
		self.length = len (self.stream)
		self.pos = 0
		self.tokens = []
		self.token = ""
		self.field = None
		self.lparens = 0
		self.rparens = 0
	
	def addParen (self, pa):
		if pa == LPAREN:
			self.lparens += 1
			
		elif pa == RPAREN:
			self.rparens += 1
			if self.rparens > self.lparens:
				self.rparens -= 1
				return
		
		if pa == LPAREN:
			if not self.tokens:
				self.tokens.append (LPAREN)
				return
				
			if self.tokens [-1] == RPAREN or self.tokens [-1] not in (PLUS,MINUS,TIMES,DIVIDE):
				self.addOperator (TIMES)
				self.tokens.append (LPAREN)
				return
			
			self.tokens.append (LPAREN)	
				
		elif pa == RPAREN:
			if not self.tokens:
				self.rparens -= 1
				return
			
			if self.tokens [-1] in (LPAREN,):
				del self.tokens [-1]
				self.lparens -= 1
				self.rparens -= 1
				return # no token in parens
			
			if self.tokens [-1] in (PLUS,MINUS,TIMES,DIVIDE):
				del self.tokens [-1]	
			
			self.tokens.append (RPAREN)
			
	def addOperator (self, op):
		if not self.tokens and op == MINUS: 
			raise ValueError #disallowed - operator
		elif not self.tokens: 
			return	
			
		if self.tokens [-1] == LPAREN: return
		if self.tokens [-1] in (PLUS,MINUS,TIMES,DIVIDE):
			if op == TIMES: return # maybe space
			self.tokens [-1] = op # than replace new operand
		else:				
			self.tokens.append (op)
			
	def addToken (self):
		self.token = self.token.strip ()
		if not self.token: return
		if self.token in SOP:
			self.addOperator (SOP [self.token])
			self.token = ""
			return
		
		if self.tokens and self.tokens [-1] not in (PLUS,MINUS,TIMES,DIVIDE):
			self.addOperator (TIMES)
			
		if self.field:
			if self.field [:3] == "all":
				self.tokens.append (self.field [3:] + ":" + self.token)
			elif self.field [-4:] == "_all":	
				self.tokens.append (self.field [-4:] + ":" + self.token)
			else:
				self.tokens.append (self.field + ":" + self.token)	
		else:	
			self.tokens.append (self.token)
				
		self.token = ""
	
	def setField (self):
		self.field = self.token
		self.token = ""
	
	def unsetField (self):
		self.field = ""
	
	def readPhrase (self):
		self.token = "`"
		while self.pos < self.length:
			c = self.stream [self.pos]
			if c == '\\':
				if self.pos < self.length - 1 and self.stream [self.pos + 1] == '"':
					self.token += '"'
					self.pos += 2
				else:	
					self.token += '\\'
					self.pos += 1					
				
			elif c == '"':
				self.pos += 1
				break
			
			else:	
				self.token += c		
				self.pos += 1
		
		self.addToken ()
			
	def readField (self):
		self.setField ()
		self.addParen (LPAREN)
		
		while self.pos < self.length:
			c = self.stream [self.pos]
			if c == '"' and not self.token:
				self.pos += 1
				self.readPhrase ()				
			
			elif c in ",+":
				self.addToken ()
				self.addOperator (PLUS)
				self.pos += 1
			
			elif c in " ()":
				self.addToken ()
				break
									
			else:
				self.token += c
				self.pos += 1
			
		self.addToken ()
		if self.field [:3] != "all" and self.field [-4:] != "_all":
			self.unsetField ()
			
		self.addParen (RPAREN)
					
	def tokenize (self):
		while self.pos < self.length:		
			c = self.stream [self.pos]
			if c == '"':
				self.pos += 1
				self.readPhrase ()
			
			elif c == ':':
				if self.token:
					self.pos += 1
					self.readField ()
				else:
					self.pos += 1
					self.addToken ()				
			
			elif c == "/":
				if self.token:
					self.token += c
				self.pos += 1	
					
			elif c in "-+":
				if self.pos == 0 or self.stream [self.pos - 1].isspace ():
					self.pos += 1
					self.addToken ()
					if c == "+": 
						self.addOperator (PLUS)
					else:
						self.addOperator (MINUS)
				else:
					self.token += c
					self.pos += 1
		
			elif c in "()":
				self.pos += 1
				self.addToken ()
				self.addParen (c)
				self.unsetField ()
			
			elif c == " ":
				self.pos += 1
				self.addToken ()
				self.addOperator (TIMES)
				
			else:
				self.token += c
				self.pos += 1
		
		# finally
		self.addToken ()
		while 1:
			if not self.tokens: break
			if self.tokens [-1] in (LPAREN,PLUS,MINUS,TIMES,DIVIDE):
				self.tokens = self.tokens [:-1]
			else:
				break	
		
		for i in range (self.lparens - self.rparens):
			self.tokens.append (RPAREN)		
		
		self.tokens.append (EOF)
		return self.tokens


class TermLexer (QueryLexer):	
	def addToken (self):
		self.token = self.token.strip ()		
		if not self.token: return
		
		self.tokens.append (self.token)				
		self.token = ""
				
	def addNear (self):
		if self.tokens and self.tokens [-1].find ("^") != -1:
			return
			
		self.tokens.append (self.token)
		self.token = ""
		
	def readNear (self):
		self.token = "^"
		while self.pos < self.length:
			c = self.stream [self.pos]
			if c.isdigit ():
				self.token += c
				self.pos += 1
			
			else:
				break
		
		if len (self.token) > 1:
			self.addNear ()
		else:
			self.token = ""	
		
	def tokenize (self):
		while self.pos < self.length:
			c = self.stream [self.pos]
			if c == "^":
				self.addToken ()
				self.pos += 1
				self.readNear ()
			
			elif c.isalnum () or c in "-/+#*|" or ord (c) > 127:
				self.token += c
				self.pos += 1
			
			else:
				self.addToken ()
				self.pos += 1
		
		self.addToken ()
		self.tokens = [x.strip () for x in self.tokens]
		self.tokens = [x for x in self.tokens if x not in (PLUS,MINUS,TIMES,DIVIDE)]
		
		while 1:
			if not self.tokens: break
			if self.tokens and self.tokens [-1].find ("^") != -1:
				self.tokens = self.tokens [:-1]
			else:
				break
						
		return self.tokens
	
if __name__ == "__main__":
	f = QueryLexer (
		'home:56.14343/-4.1~10.45'
	)
	
	print(f.tokenize ())
	
			
		
	


from delune import _delune
from delune import util
from datetime import datetime
from . import queryparser
import types
import copy
import re
import time
import fieldreader
import typeinfo
import memory

class Query:
	DEBUG = 0
	CACHEPAGE = 3
	def __init__ (self, segments, qs, offset, fetch, sort, analyze):
		self.segments = segments
		self.qs = qs
		self.offset = offset
		self.fetch = fetch
		self.sort = sort		
		self.analyze = analyze
		
		self.cachedfetch = self.offset + self.fetch
		if self.offset == 0 and len (self.segments) < self.CACHEPAGE:
			if len (self.segments):
				self.cachedfetch = self.cachedfetch + int ((float (self.fetch) / len (self.segments)))
			else:
				self.cachedfetch = 0
				
		self.sorted = 0		
		# postential sort options
		self.sortorder = 0
				
		if self.sort:
			if self.sort [0] == "-":
				self.sortorder = 1
				self.sort = self.sort [1:]
			else:
				self.sortorder = -1
		
		self.hits = _delune.Compute (memory.get ())
		
		self.sortindex = None
		self.totalcount = 0		
		self.duration = 0
		self.esegments = []
		self.bucket = []
		self.highlights = {}
	
	def getBufferSize (self, segment, size):		
		hits = self.hits.count ()
		if hits == 0: return 0
		bsize = int ((self.segments.si.getSegmentNumDoc (segment.seg) / float (hits)) * size)
		if bsize > memory.get_buffer_size (): bufsize = 1024
		else: bufsize = 0		
		return bufsize

	def close (self):
		self.hits.close ()		
	
	def setResult (self, cached):
		self.totalcount, highlights, self.bucket = cached
		for each in highlights:
			self.highlights [each] = None
		
	def newSegmentScan (self, segment):
		self.randomscan = 0
		self.numoperate = 0
		self.numnorm = 0
		self.mem = memory.get (segment)
		self.fi = fieldreader.FieldReader (self.mem, segment, self.analyze and self.segments.inf.analyzer or None, self.sort == "tfidf")
		self.hits.setN (self.sort == "tfidf" and self.segments.si.getSegmentNumDoc (segment.seg) or 0)		
		
	def addHighlight (self, term):
		self.highlights [term] = None
		if len (term) < 4: return
		
		# plural
		if term [-1] != 's':
			if term [-1] in "xz":
				self.highlights [term + 'es'] = None
			elif term [-1] == 'y':
				if term [-2] in "aeiou":
					self.highlights [term + "s"] = None
				else:	
					self.highlights [term [:-1] + "ies"] = None			
			elif term [-1] == 'f':
				self.highlights [term [:-1] + "ves"] = None
				self.highlights [term + "s"] = None
			elif term [-2:] == 'fe':
				self.highlights [term [:-2] + "ves"] = None	
				self.highlights [term + "s"] = None	
			else:
				self.highlights [term + 's'] = None
				
		elif len (term) >= 5 and term [-3:] == 'ies':
			self.highlights [term [:-3] + "y"] = None
		
		elif len (term) >= 5 and term [-3:] == 'ves':
			self.highlights [term [:-3] + "f"] = None
			self.highlights [term [:-3] + "fe"] = None
				
		elif len (term) >= 5 and term [-2:] == 'es' and (term [-3] in "sxz" or term [-4:-2] in ("sh", "ch")):
			self.highlights [term [:-2]] = None
			self.highlights [term [:-1]] = None
		
		elif term [-1] == 's':
			self.highlights [term [:-1]] = None
		
	def setTi (self, term):
		try: 
			name, value = term.split (":")
		except ValueError:
			name, value = "default", term
			
		type, num = self.segments.si.getFieldInfo (name)		
		return self.fi (type, term, name, value, num)		
		
	def optimize (self, segment, expression):
		#swaping operands
		# it reduces push-pop operation
		if type (expression [1]) is bytes:
			expression [1] = self.setTi (expression [1])			
								
		if type (expression [2]) is bytes:
			expression [2] = self.setTi (expression [2])			
		
		#swap between operands
		if expression [0] in "*+":
			if expression [2] is None and self.hits.count ():
				expression  = [expression [0], None, expression [1]]
				if self.DEBUG: print("SWAP:", expression)				
			
			# swap phrase		
			if expression [1] and expression [2]:
				if not expression [1].methology == "Phrase" and expression [2].methology == "Phrase":
					expression  = [expression [0], expression [2], expression [1]]
					if self.DEBUG: print("SWAP:", expression)				
			
				#swap with df	
				if expression [1].ti and expression [2].ti:
					if expression [1].ti [0] > expression [2].ti [0]:
						expression = [expression [0], expression [2], expression [1]]
						if self.DEBUG: print("SWAP:", expression)				
			
				#swap range						
				if typeinfo.typemap.getsize (expression [1].type) and not typeinfo.typemap.getsize (expression [2].type):
					expression = [expression [0], expression [2], expression [1]]
					if self.DEBUG: print("SWAP:", expression)
					
		return expression
	
	rx_proxdelta = re.compile ("\s*(?P<delta>\^[0-9]+)\s*")
	def computePhrase (self, segment, term, operator):
		if not term.type: return 0
		if not term.value: return 0
		dc = 0
		self.randomscan = 0
		
		if len (term.value) == 1: # not phrase
			term.value [0].readprox = 0
			return self.computeTerm (segment, term.value [0], operator)
						
		pushed = 0
		if self.hits.count ():
			self.hits.push ()
			pushed = 1
			if self.DEBUG: print("P-PUSH:", self.hits.saved ())
			
		for termTi in term.value:
			if not termTi.df: break			
			self.computeTerm (segment, termTi, operator)
			dc = self.hits.intersect (termTi.near, termTi.loose)
			if not dc: break
		
		if pushed:
			self.hits.pop ()
			if self.DEBUG: print("P-POP:", self.hits.saved ())
			
		return dc
		
	def computeWildCard (self, segment, term, operator):
		if not term.ti: return 0
		
		self.randomscan = 0
		pushed = 0
		if self.hits.count ():
			self.hits.push ()
			pushed = 1
			if self.DEBUG: print("P-PUSH:", self.hits.saved ())
		
		for termTi in term.value:
			if not self.computeTerm (segment, termTi, operator): continue
			dc = self.hits.union ()
		
		if pushed:
			self.hits.pop ()
			if self.DEBUG: print("P-POP:", self.hits.saved ())			
		
		return dc
				
	def computeTerm (self, segment, term, operator):
		if not term.type: return 0
		if term.stopword: return 0
		if not term.df: return 0
		
		if segment.rd: self.randomscan = 0
		
		if self.randomscan and len (self.bucket) >= self.cachedfetch:
			if self.DEBUG: print("SKIP: randomscan")
			self.hits.setcount (term.df)
			return term.df
			
		df, doff, poff, skip, plen = term.ti
		if (term.sp): segment.sm.read (self.mem, self.getBufferSize (segment, term.sp [1]), *term.sp) # needn't for f type
		
		if self.randomscan:
			if self.DEBUG: print("RANDOM SCAN")
			segment.readPosting (self.mem, df, doff, poff, skip, plen, 0, self.cachedfetch, term.readprox)
			self.hits.setcount (term.df)
			
		else:
			if self.DEBUG: print("FULL SCAN")
			segment.readPosting (self.mem, df, doff, poff, skip, plen, -1, -1, term.readprox)			
			
		if operator != "-" and typeinfo.typemap.hasnorm (term.type):
			self.numnorm += 1 # need sort?
			if term.query:		
				for hk in term.query.split ():
					self.addHighlight (hk)
		
		t =  self.hits.set (self.segments.si.getWeight (term.name))
		return t
		
	def computeCoord (self, segment, term, operator):
		if not term.value: return 0
		latitude, longitude, distance = term.value
		set = 0
		if self.sort == term.name: set = 1
		
		segment.sm.read (self.mem, self.getBufferSize (segment, term.sp [1]), *term.sp)
		return self.hits.distance (latitude, longitude, distance, set)
	
	def computeDigit (self, segment, term, operator):
		if not term.value: return 0
		
		a, b = term.value
		if b in ("all", "any", "none"):
			smfunc = self.hits.bit			
		else:
			smfunc = self.hits.between
		
		segment.sm.read (self.mem, self.getBufferSize (segment, term.sp [1]), *term.sp)
		
		if not (a == -1 and (b == -1 or b in ("all", "any", "none"))):
			self.randomscan = 0
				
		if self.randomscan:
			wholes = segment.getNumDoc ()  - segment.getDeletedCount ()
			self.hits.setcount (wholes)			
			
			if len (self.bucket) >= self.cachedfetch:
				return wholes
				
			dc = smfunc (a, b, operator, self.cachedfetch + segment.getDeletedCount ())
			return wholes
			
		else:
			return smfunc (a, b, operator, -1)
				
	def compute (self, segment, expression):
		if self.DEBUG: print("EXPRESSION", expression)
		
		record_count = 0
		operate = 1
		
		#---------------------------------------------------------------------------------
		# Initial Optimazation
		#---------------------------------------------------------------------------------	
		if type (expression) is bytes: # only 1 term
			expression = ["*", expression, None]
			operate = 0
			
		elif expression [0] != "-" and type (expression [1]) == bytes and type (expression [2]) == list:
			expression = [expression [0], expression [2], expression [1]]		
		
		#---------------------------------------------------------------------------------
		# Recursive Compute
		#---------------------------------------------------------------------------------	
		if type (expression [1])  is list:
			expression [1] = self.compute (segment, expression [1])
	
			if self.numoperate and expression [1] is None and self.hits.count () == 0 and expression [0] in "*-":
				if self.DEBUG: print("SKIP:", expression [2])
				self.hits.abort () # need not proceed
				expression [2] = None
				return None
		
		if type (expression [2])  is list:
			expression [2] = self.compute (segment, expression [2])

			if self.numoperate and expression [2] is None and self.hits.count () == 0 and expression [0] in "*":
				if self.DEBUG: print("SKIP:", expression [1])				
				self.hits.abort () # need not proceed
				expression [1] = None
				return None
	
		#---------------------------------------------------------------------------------
		# Optimize
		#---------------------------------------------------------------------------------		
		expression = self.optimize (segment, expression)
		
		#---------------------------------------------------------------------------------
		# Stack Mangement
		#---------------------------------------------------------------------------------	
		# done one session, pop
		if (expression [1] is None and expression [2] is None):
			self.hits.pop ()
			if self.DEBUG: print("POP:", self.hits.saved ())
		
		# new operand, if count() save prev results
		elif expression [1] and self.hits.count ():
			self.hits.push ()			
			if self.DEBUG: print("PUSH:", self.hits.saved ())
			
		#---------------------------------------------------------------------------------
		# Left Operand
		#---------------------------------------------------------------------------------		
		# operation count			
		if  expression [1] and not expression [1].stopword:			
			self.numoperate += 1
			if type (expression [1]) is not list:
				expression = self.optimize (segment, expression)								
			# left operand is always "*"
			record_count = getattr (self, "compute" + expression [1].methology) (segment, expression [1], "*")
			if self.DEBUG: print("QUERY:", expression [1], record_count)
			if not record_count	and expression [0] in "-*":
				self.hits.abort () # need not proceed
				return None
		
		#---------------------------------------------------------------------------------
		# Right Operand
		#---------------------------------------------------------------------------------		
		if  expression [2] and not expression [2].stopword:
			self.numoperate += 1
			expression = self.optimize (segment, expression)
			record_count = getattr (self, "compute" + expression [2].methology) (segment, expression [2], expression [0])			
			if self.DEBUG: print("QUERY:", expression [2], record_count)			
			if not record_count:
				if expression [0] in "*":
					self.hits.abort () # + or - is keep hits
					return None		
		
		#---------------------------------------------------------------------------------
		# Operate
		#---------------------------------------------------------------------------------		
		if self.DEBUG: 
			print("START OPERATE:", expression, record_count)
		
		if operate:
			if expression [0] == "*":
				record_count = self.hits.intersect ()
			elif expression [0] == "+":	
				record_count = self.hits.union ()				
			elif expression [0] == "-":	
				record_count = self.hits.trim ()
			
			if self.DEBUG: print("OPERATE:", expression, record_count)
		
		if self.DEBUG: print("CURSAVED:", self.hits.saved ())		
		return None
			
	def merge (self):
		if self.sortindex:
			if self.sorted == 1:
				self.bucket.sort (lambda x, y: cmp (x [self.sortindex], y [self.sortindex]))
			elif self.sorted == -1:
				self.bucket.sort (lambda x, y: cmp (y [self.sortindex], x [self.sortindex]))
		
		highlights = list(self.highlights.keys ())
		highlights.sort (lambda x, y: cmp (len (y), len (x)))
		#rx_highlights  = re.compile ("(^| )(?P<match>" + "|".join (highlights) + ")($| )", re.I)
			
	def query (self):
		s = time.time ()
		
		tokens = queryparser.Tokenizer (self.qs).tokenize ()	
		
		try:
			expression = queryparser.ArithmeticParser(tokens).parse_expression()
		except IndexError:
			self.close ()
			return [501, 0, 0, ""]	
		
		for segment in self.segments:	
			try:
				self.newSegmentScan (segment)			
				# 1 term and no sort and has freq data
				if  not self.sort and type (expression) is bytes:
					self.randomscan = 1
					
				clone = copy.deepcopy (expression)			
				self.query_segment (segment, clone)
			
			except IOError:
				self.esegments.append (segment.seg)
				self.segments.si.trace ()				
			
		if self.DEBUG: print("Total Result:", self.totalcount)		
		
		self.duration = int ((time.time () - s) * 1000) #ms	
		self.merge ()
		
	def query_segment (self, segment, expression):
		#-------------------------------------------------------------------------------
		# operate
		#-------------------------------------------------------------------------------
		self.compute (segment, expression)
		
		#-------------------------------------------------------------------------------
		# skip extra fetching when no sort
		#-------------------------------------------------------------------------------
		if not self.sort and len (self.bucket) > self.cachedfetch:
			count = self.hits.count ()
			self.totalcount += count
			return
			
		#-------------------------------------------------------------------------------
		# sort setting
		#-------------------------------------------------------------------------------		
		bysortkey = 0
		byextra = 0		
		if self.sort == "tfidf" and self.numnorm:
			self.sorted = -1
			self.sortindex = -1
		
		elif self.sort:
			type, num = self.segments.si.getFieldInfo (self.sort)
			if typeinfo.typemap.isint (type):
				sp = segment.getSortMapPointer (num)					
				if sp:
					pointer, vsize = sp
					segment.sm.read (self.mem, self.getBufferSize (segment, vsize), pointer, vsize)
					bysortkey = 1
					self.sorted = self.sortorder
					self.sortindex = -1					
			
			elif typeinfo.typemap.isextra (type):
				byextra = 1
				self.sorted = self.sortorder
				self.sortindex = -2
				
		else:
			self.sorted = 0
			self.sortindex = None
			
		want = self.cachedfetch
		self.hits.sort (want, self.sorted, bysortkey, byextra)
		
		#-------------------------------------------------------------------------------
		# select from hit docs
		#-------------------------------------------------------------------------------
		bucket = self.hits.hitdoc (segment.seg, want, segment.getBits ())		
		self.bucket =  bucket + self.bucket
		
		#-------------------------------------------------------------------------------
		# counting hitdoc
		#-------------------------------------------------------------------------------
		count = self.hits.count ()
		self.totalcount += count
				
		if self.DEBUG: print("Segment Result:", count)
		
	


#!/usr/bin/env python
import os
import glob
from distutils.core import setup, Extension
import platform
osbit, _dummy = platform.architecture()
dira = osbit == "32bit" and "x86" or "x64"
data_files = []

if os.name == "nt":
	include_dirs = ["../win32/__prebuilts__/2.7/zlib/%s" % dira, "../win32/__prebuilts__/2.7/pthread2/%s" % dira]
	library_dirs = ["../win32/__prebuilts__/2.7/zlib/%s" % dira, "../win32/__prebuilts__/2.7/pthread2/%s" % dira]
	libraries = ["zlib", "pthreadVC2"]	
else:
	include_dirs = ["/usr/include"]
	library_dirs = ["/usr/lib/x86_64-linux-gnu"]
	libraries = ["z", "pthread"]
		
module = Extension(
	'_delune',
	sources = [
		'core.c', 
		'analyzer/analyzer.c', 		
		'analyzer/stopword.c', 
		'analyzer/endword.c', 
		'analyzer/formalizer.c', 		
		'analyzer/stem.c',
		'analyzer/stem_fr.c',
		'analyzer/stem_de.c',
		'analyzer/stem_it.c',
		'analyzer/stem_es.c',
		'analyzer/stem_pt.c',
		'analyzer/stem_fi.c',
		'analyzer/stem_sv.c',
		'analyzer/stem_hu.c',
		'analyzer/stem_ar.c',
		'analyzer/removeaccents.c',
		'index/sort.c', 
		'index/osutil.c', 
		'index/search.c', 
		'index/compfunc.c',
		'index/heapsort.c', 
		'index/termhashtable.c', 
		'index/generichash.c',
		'index/bfile.c', 
		'index/zip.c',
		'index/mempool.c',
		'index/ibucket.c',
		'index/util.c',
		'mod_analyzer.c',		
		'mod_util.c', 
		'mod_int.c', 
		'mod_bits.c', 
		'mod_mergeinfo.c', 
		'mod_document.c',
		'mod_posting.c', 
		'mod_termtable.c',
		'mod_terminfo.c',
		'mod_sortmap.c',   
		'mod_memorypool.c',		
		'mod_selector.c',		
		'mod_classifier.c',
		'mod_dbint.c',
		'mod_calculator.c',
		'mod_binfile.c',		
		'mod_sgmlparser.c',
		'mod_compute.c',
	],
	include_dirs = include_dirs,
	library_dirs = library_dirs,
	libraries = libraries
)

setup (name = '_delune',
       version = '2.0.0',       
       author = 'Hans Roh',
       author_email = 'hansroh@gmail.com',
       url = '',
       ext_modules = [module]       
)




from delune import _delune
import os, glob
import pickle as pk
import zlib

class BinFile:
	def __init__ (self, path, mode = "r"):
		self.path = path
		self.mode = mode		
		
		if mode == "r":
			flag = os.O_RDONLY
		else:
			flag = os.O_WRONLY | os.O_CREAT			
		if os.name == "nt":
			flag |= os.O_BINARY
		
		oldmask = os.umask (0o113)
		self.fdno = os.open (self.path, flag)
		os.umask (oldmask)
		self.bf = _delune.BinFile (self.fdno, self.mode)		
	
	def __enter__ (self):
		return self
	
	def __exit__ (self, type, value, tb):
		self.close ()
			
	def __getattr__ (self, attr):
		return getattr (self.bf, attr)
		
	def close (self):
		self.bf.close ()
		os.close (self.fdno)


class ZPickled:
	def __init__ (self, filename, mode="r", size = 1000000000):
		self.filename = filename
		self.mode = mode
		self.size = size
		self.colname = os.path.basename (filename)
		self.fileindex = 0
		self.reopens = 0
		self.count = 0
		self.pos = 0
		self.open ()
	
	def __enter__ (self):
		return self
	
	def __exit__ (self, type, value, tb):
		self.close ()
			
	def __len__ (self):
		g = glob.glob (self.filename + "*")
		g.sort ()
		db = BinFile (g [-1], "r")
		c = db.readInt4 ()
		return c		
		
	def open (self):
		if self.mode == "a":
			for i in range (1000):
				if os.path.isfile ("%s.%d" % (self.filename, i)):
					self.fileindex = i
					self.reopens = 1
					break					
			if self.reopens:
				fn = "%s.%d" % (self.filename, self.fileindex)
				self.fileindex += 1
			else:
				fn =  self.filename
			self.db = BinFile (fn, "r")
			self.count = self.db.readInt4 ()
			self.close ()
			self.db = BinFile (fn, "a")
		
		else:
			self.db = BinFile (self.filename, self.mode)
			if self.mode == "w": 
				self.db.writeInt4 (0)
			elif self.mode == "r": 
				self.count = self.db.readInt4 ()			
	
	def reopen (self):
		self.close ()
		if self.mode == "r":
			if os.path.isfile ("%s.%d" % (self.filename, self.fileindex)):
				self.db = BinFile ("%s.%d" % (self.filename, self.fileindex), "r")
				self.count = self.db.readInt4 ()
		else:
			self.db = BinFile ("%s.%d" % (self.filename, self.fileindex), "w")
			if self.mode in "wa": self.db.writeInt4 (0)							
		self.fileindex += 1
		self.reopens += 1
	
	def __iter__ (self):
		return self
	
	def __next__ (self):
		d = self.get ()
		if not d:
			raise StopIteration
		return d	
		
	def deserialize (self, d):
		return pk.loads (zlib.decompress (d))
		
	def serialize (self, d):
		return zlib.compress (pk.dumps (d, 1), 9)
	
	def close (self):
		if self.mode in "wa":
			self.db.seek (0)
			self.db.writeInt4 (self.count)
		
		if self.db:
			self.db.close ()
			self.db = None

	def truncate (self):
		if self.mode == "w" and os.path.isfile (self.filename):
			self.close ()
			os.remove (self.filename)
			self.open ()
	
	def write (self, post):		
		self.db.writeBytes (self.serialize (post))
	
	def get (self):	
		if self.pos == self.count:
			self.reopen ()
			if self.pos == self.count: 
				return None
		self.pos += 1
		return self.deserialize (self.db.readBytes ())
	
	def add (self, post):	
		self.addmany ([post])
		
	def addmany (self, posts):
		for post in posts:
			self.write (post)
			self.count += 1
			if self.db.tell () > self.size:
				self.reopen ()

def open (path, mode = "r"):
	return BinFile (path, mode)

def openzp (path, mode = "r", size = 1000000000):
	return ZPickled (path, mode, size)



import delune
from delune import binfile
from delune.lib import logger as loggerlib
import os, codecs, glob, json

resource_dir = None

def add_document (indexer, doc):
	document = delune.document ()	
	document.documents (doc ['documents'])
	if 'snippet' in doc:
		document.snippet (*tuple (doc ['snippet']))
	for name, opt in doc ['fields'].items ():
		if name == "_id":			
			data = str (opt)
			lang = 'un'
			ftype = delune.STRING
		else:	
			data = opt.get ('data')
			lang = opt.get ('lang', 'un')
			ftype = opt.get ('type')
			if not ftype:
				raise TypeError ('Type of field is required')
		document.field (name, data, ftype, lang)
	indexer.add_document (document)

def getdir (*d):
	global resource_dir	
	return os.path.join (resource_dir, *d)

def delete (col, opt, deletables):
	global log
	
	opt ['searcher']['remove_segment'] = 0
	searcher = col.get_searcher (**opt ['searcher'])
	if searcher.si.getN () == 0:
		searcher.close ()
		return []
	
	for qs in deletables:
		qs ["commit"] = False
		searcher.delete (**qs)
		
	searcher.commit ()	
	searcher.close ()	
	log ('delete %d documents' % len (deletables))	
	return []
		
def index (path):	
	global log
	
	alias = os.path.split (path) [-1]
	if alias [0] in "#-":
		return
		
	with codecs.open (path, "r", "utf8") as f:
		colopt = json.loads (f.read ())
	
	colopt ['data_dir'] = [getdir (os.path.normpath(d)) for d in colopt ['data_dir']]
	queue_dir = colopt ['data_dir']
	if type (queue_dir) is list:
		queue_dir = queue_dir [0]
	
	analyzer = delune.standard_analyzer (3000, 1, **colopt ['analyzer'])
	col = delune.collection (
	  indexdir = colopt ['data_dir'],
	  mode = delune.APPEND,
	  analyzer = analyzer,
	  version = colopt.get ("version", 1)
	)
	
	queue_dir = os.path.join (queue_dir, ".que")
	queues = [(os.path.basename (q), os.path.getmtime (q)) for q in glob.glob (os.path.join (queue_dir, 'que.*')) if not q.endswith (".lock")]
	if not queues:
		log ("nothing to for %s in %s" % (os.path.split (path) [-1], os.path.dirname (path)))
		col.removeDeletables ()
		return
	
	# reverse sort for trucate
	queues.sort (key = lambda x: x [-1], reverse = True)
	deletables = []
	
	index = 0
	for gfile, mtime in queues:
		if gfile == "que.truncate":
			break
		index += 1
	
	if index < len (queues):
		log ("%s found trcuate request" % alias)
		for qfile, mtime in queues [index:]:
			os.remove (os.path.join (queue_dir, qfile))
		queues = queues [:index]
		indexer = col.get_indexer (**colopt ['indexer'])
		indexer.truncate ()	
		indexer.close ()
	
	deleted = 0
	# resort from old one
	queues.reverse ()	
	for qfile, mtime in queues:
		bf = binfile.BinFile (os.path.join (queue_dir, qfile), "r")
		while 1:
			try: cmd = bf.readVInt ()
			except OSError:
				bf.close ()
				break
			
			try:
				doc = json.loads (bf.readZBytes ().decode ("utf8"))
			except (OSError, MemoryError):
				bf.close ()
				break
			
			if cmd == 1:
				qs = doc.get ("query")				
				deleted += 1
				deletables.append (qs)				
				if len (deletables)	== 1000:
					deletables = delete (col, colopt, deletables)
	
	if deletables:
		deletables = delete (col, colopt, deletables)	
	log ('%d documents deleted in this session' % len (deletables))	
			
	indexer = col.get_indexer (**colopt ['indexer'])	
	# mannual commit mode, ignore memory usage
	indexer.set_autocommit (False)
	
	indexed_docs = 0
	indexed_ques = []
	
	for qfile, mtime in queues:
		bf = binfile.BinFile (os.path.join (queue_dir, qfile), "r")
		while 1:
			try: cmd = bf.readVInt ()				
			except OSError:
				bf.close ()
				break
			
			if cmd == 1:	 # just delete
				bf.readZBytes ()
				continue
			
			try:
				doc = json.loads (bf.readZBytes ().decode ("utf8"))				
			except (OSError, MemoryError):
				bf.close ()
				break
			
			indexed_docs += 1	
			add_document (indexer, doc)
		
		indexed_ques.append (os.path.join (queue_dir, qfile))		
		if maybe_commit (indexer, indexed_ques):
			indexed_docs = 0
			indexed_ques = []
	
	maybe_commit (indexer, indexed_ques, force = True)	
	indexer.close (optimize = deleted)

def maybe_commit (indexer, indexed_ques, force = False):
	if force or indexer.is_memory_over ():
		# commit and delete queue for preventing duplication
		indexer.commit ()
		for qpath in indexed_ques:
			log ('remove indexed que %s' % os.path.basename (qpath))
			os.remove (qpath)
		return 1
	return 0

			
if __name__ == "__main__":
	import sys, getopt
	
	log = loggerlib.screen_logger ()
	argopt = getopt.getopt(sys.argv[1:], "l:", ["log="])

	for k, v in argopt [0]:
		if k == "-l" or k == "--log":
			log = loggerlib.multi_logger ([log, loggerlib.rotate_logger (v, "delune", "daily")])
			
	delune.configure (1, log)
	try:
		resource_dir = argopt [1][0]
	except IndexError:
		print ('Usage: mirror.py resource_dir')	
		print ('  ex. mirror.py /var/tmp/delune')	
		sys.exit ()
		
	conf_dir = os.path.join (resource_dir, "models", ".config")
	for alias in os.listdir (conf_dir):
		log ("starting index for %s" % alias)
		try:
			index (os.path.join (conf_dir, alias))
		except:
			log.trace ()			
		log ("index done for %s" % alias)
		


import threading
import os
from delune import _delune

class Memory:
	def __init__ (self, threads, buffer, limit, scope = "disk", logger = None):
		self.mpool = _delune.MemoryPool (threads, buffer, int (limit / float (threads)))
		self.limit = limit * (1024 << 10)
		self.threads = threads
		self.scope = scope
		self.logger = logger
		self.mutexes = {}
		self.lock = threading.Lock ()
		self.thread_id_map = {}
		self.got_main_thread = 0
	
	def usage (self):
		u = []
		for i in range (self.threads):
			bytes = self.mpool.usage (i)
			u.append ((i, bytes))
		return u	
	
	def maintern (self):
		# maintern memory
		if self.mpool.usage () > self.limit:
			ousage = self.mpool.usage ()
			self.mpool.maintern (self.__gettid ())
			self.logger ("memory reduced %d -> %d" % (ousage, self.mpool.usage ()), "info", "")		

	def recover (self):
		self.mpool.recover (self.__gettid ())

	def close (self):
		self.mpool.close ()

	def __getscope (self, path):
		if self.scope == "disk":
			return os.path.split (os.path.split (os.path.normpath (path)) [0]) [0]
		elif self.scope == "index":
			os.path.split (os.path.normpath (path)) [0]
		else:
			return os.path.normpath (path)

	def __gettid (self):
		tid = id (threading.currentThread ())
		try: 
			return self.thread_id_map [tid]
		except KeyError:
			if (threading.currentThread ().getName() == "MainThread"):
				self.thread_id_map [tid] = 0
				self.got_main_thread = 1
				return 0
			tids_count = len (self.thread_id_map) - self.got_main_thread
			if tids_count == self.threads:
				raise SystemError("Too Many Threads")				
		self.thread_id_map [tid] = tids_count
		return tids_count

	def set (self, segment):
		if self.scope == "global": return -1

		scope = self.__getscope (os.path.join (segment.home, str (segment.seg)))
		self.lock.acquire ()
		try:
			mutex_id = self.mutexes [scope]
		except KeyError:
			mutex_id = self.mpool.newmutex ()
			self.mutexes [scope] = mutex_id

		self.lock.release ()
		return mutex_id

	def get (self, segment = None):
		if segment is None:
			return self.mpool.get (self.__gettid (), -1)
		return self.mpool.get (self.__gettid (), segment.mutex_id)

	def remove (self, segment):
		if self.scope != "segment": return 0

		mutex_id = segment.mutex_id
		if mutex_id == -1: return 0

		self.lock.acquire ()
		for scope, _mutex_id in list(self.mutexes.items ()):
			if _mutex_id == mutex_id:
				del self.mutexes [scope]
				self.mpool.delmutex (mutex_id)
		self.lock.release ()
		return 1

memory = None

def isInitialized ():
	return memory
	
def initialize (threads, buffer, limit, mutexscope = "disk", logger = None):
	global memory
	memory = Memory (threads, buffer, limit, mutexscope, logger)
	
def set (segment):
	global memory
	if memory:
		return memory.set (segment)	

def get (segment = None):
	global memory
	if memory:
		return memory.get (segment)	

def remove (segment):
	global memory
	if memory:
		memory.remove (segment)
	
def destroy ():
	global memory
	if memory:
		memory.close ()
		memory = None
	
def recover ():
	global memory
	if memory:
		memory.recover ()
	
def get_buffer_size ():
	global memory
	if memory:
		return memory.mpool.buffer_size	

def maintern ():
	global memory
	memory.maintern ()

def usage ():
	global memory
	return memory.usage ()
	


# standard

typeinfo = {	
	"Text": {"method": "Text"},
	"Term": {"method": "Text"},
	
	"String": {"method": "Field"},
	"List": {"method": "List"},

	"Coord4": {"len": 6, "method": "Coord"},
	"Coord6": {"len": 8, "method": "Coord"},
	"Coord8": {"len": 10, "method": "Coord"},
	
	"Bit8":  {"len": 1, "method": "Bit"},
	"Bit16": {"len": 2, "method": "Bit"},
	"Bit24": {"len": 3, "method": "Bit"},
	"Bit32": {"len": 4, "method": "Bit"},
	"Bit40": {"len": 5, "method": "Bit"},
	"Bit48": {"len": 6, "method": "Bit"},
	"Bit56": {"len": 7, "method": "Bit"},
	"Bit64": {"len": 8, "method": "Bit"},
	
	"Int8":  {"len": 1, "method": "Int"},
	"Int16": {"len": 2, "method": "Int"},
	"Int24": {"len": 3, "method": "Int"},
	"Int32": {"len": 4, "method": "Int"},
	"Int40": {"len": 5, "method": "Int"},
	"Int48": {"len": 6, "method": "Int"},
	"Int56": {"len": 7, "method": "Int"},
	"Int64": {"len": 8, "method": "Int"}
}
	
class FieldType:
	def hasnorm (self, type):
		return type in ("Term", "Text")

	def isscorable (self, type):
		return self.hasnorm (type) or self.isfield (type) or self.islist (type)

	def isfield (self, type):
		return type == "String"
	
	def islist (self, type):
		return type == "List"
		
	def isbit (self, type):
		return type [:3] == "Bit"
	
	def isint (self, type):
		return type [:3] == "Int"
	
	def isextra (self, type):
		return type [:5] == "Coord"
	
	def iscoord (self, type):
		return type [:5] == "Coord"
			
	def valid (self, type):
		return type in typeinfo
	
	def getdelim (self):
		try:
			return typeinfo [type]["delim"]
		except KeyError:
			return
		
	def getsize (self, type):
		try:
			return typeinfo [type]["len"]
		except KeyError:
			return 0
	
	def getmethod (self, type):
		return typeinfo [type]["method"]
		
	
typemap = FieldType ()




from . import segmentreader
from ... import memory

class SegmentReaders:
	def __init__ (self, si):
		self.si = si
		self.numdel = 0
		self.segment_readers = []
		self.segment_map = {}
	
	def __len__ (self):
		return len (self.segment_readers)

	def __iter__ (self):
		return self.segment_readers.__iter__ ()
	
	def __getitem__ (self, i):
		return self.segment_map [int (i)]
	
	def first (self):
		return self.segment_readers [0]
	
	def last (self):
		return self.segment_readers [-1]
			
	def close (self):
		for reader in self.segment_readers:
			reader.close ()
			memory.remove (reader)
	
	def commit (self):
		for reader in self.segment_readers:
			reader.commit_deleted ()
			
	def remove (self, segs):
		segment_readers = []
		for reader in self.segment_readers:
			if int (reader.seg) in segs:
				self.si.log ("segment error, %s removed" % reader.seg)
				reader.close ()
				memory.remove (reader)
			else:
				segment_readers.append (reader)
		self.segment_readers = segment_readers
	
	def reload_deleted (self):
		for reader in self.segment_readers:
			reader.reload_deleted ()
		
	def load (self):
		latest_segments = self.si.getSegmentList ()
		segment_readers = []
		segment_map = {}

		while 1:
			try: reader = self.segment_readers.pop (0)
			except IndexError: break
			if reader.seg in latest_segments:
				reader.reload_deleted ()
				segment_readers.append (reader)
				segment_map [int (reader.seg)] = reader

			else:
				reader.close ()
				memory.remove (reader)
		
		for seg in latest_segments:
			if seg not in segment_map:
				try:
					reader = segmentreader.SegmentReader (self.si, seg)
				except:
					self.si.log ("segment %s had been broken" % seg, "fail")
					self.si.trace ()	 # deleted by indexer?
					self.clean = 0
				else:
					self.si.log ("segment %s had been loaded" % seg)
					reader.mutex_id = memory.set (reader)
					segment_map [seg] = reader
					segment_readers.append (reader)

		segment_readers.sort (key = lambda x: int (x.seg), reverse = True)
		self.numdel = len ([_f for _f in [reader.rd for reader in segment_readers] if _f])
		self.segment_readers = segment_readers
		self.segment_map = segment_map



from .segment import segment
from delune.searcher import searcher, cache
import time
import threading
import math
from .classifiers import metaclassifier
from .segment import composedsegmentreader
import sys, os
import re
import delune
import copy
from delune.lib import unistr
from delune import filesystem

class TermInfo:
	def __init__ (self):
		pass
	
	def __repr__ (self):
		r = []
		for name, attr in list(self.__dict__.items ()):
			r.append ("%s:%s" % (name, attr))
		return "<%s>" % " ".join (r)
		
	
class Classifier (searcher.Searcher):
	def __init__ (self, si, do_init = True):
		self.si = si
		
		self.reader = composedsegmentreader.ComposedSegmentReader (self.si)			
		self.clfs = {}
		
		self.top_features = self.numQueryCache = self.si.getopt (use_features_top = 0) # use all features
		
		self.numquery = 0
		self.references = 0
		self.shutdown_level = 0
		self.maintern_time = -1
		self.closed = 1
		self.clean = 0
		self.cond = threading.Condition ()
		self.mutex = threading.Lock ()
		self.mod_time = self.si.getModfiedTime ()	
		self.has_deletable = False
		self.numQueryCache = 200
		self.classifer_opts = []
		self.cache = cache.Cache (self.numQueryCache)	
		
		self._initiallized = False		
		
		if do_init: 
			self.init ()
	
	def maintern (self):
		if time.time () - self.maintern_time > self.MAINTERN_INTERVAL and self.ismainternable ():
			self.maintern_time = time.time ()
			if self.need_refresh ():				
				self.do_refresh ()
				
	def addClassifier (self, classifier):
		self.reader.si.log ("text-classifier [%s] initializing..." % classifier)
		featureset = self.getFeatureSet (classifier == "tfidf" and "similarity" or classifier) or self.default_featureset
		if not featureset:
			self.reader.si.log ("text-classifier [%s] has no features" % classifier, "fail")
			return
			
		modulename = "delune.classifier.classifiers.%s" % classifier
		__import__ (modulename)
		clf = sys.modules [modulename].Classifier (self.reader, self.si.analyzer, featureset, self.top_features, self.si.logger)
		self.clfs [classifier] = clf
		#print (self.clfs)
	
	def getFeatureSet (self, classifier):	
		return self.reader.si.segments.features.get (classifier, None)
	
	def setopt (self, cl, **option):
		self.classifer_opts.append ((cl, option))
		if type (cl) is str:
			cl = [cl]		
		for each in cl:
			self.__setopt (each, **option)
	
	def getopt (self, cl, option):
		return self.get_classifier (cl).getopt (name)
			
	def __setopt (self, cl, **option):
		self.get_classifier (cl).setopt (**option)
		
	def get_classifier (self, name):
		return self.clfs [name]
		
	def do_close (self):
		if self.closed: return 1
		if self.reader:			
			self.reader.close ()			
			self.reader = None
		self.si.close ()
		self.si = None
		self.closed = 1

	def do_refresh (self, *arg, **karg):
		self.clean = 1
		if self.si.getModfiedTime () == -1:
			self.clean = 0
			return 
						
		try:
			self.reader.load ()
		except:
			# if failed, loading next time
			self.si.trace ()
			self.clean = 0
			return 0
		
		self.default_featureset = self.getFeatureSet ("default")
		
		self.clfs = {}
		self.addClassifier (delune.NAIVEBAYES)
		self.addClassifier (delune.FEATUREVOTE)	
		self.addClassifier (delune.ROCCHINO)
		self.addClassifier (delune.TFIDF)
		self.addClassifier (delune.SIMILARITY)		
		self.addClassifier (delune.MULTIPATH)
		self.addClassifier (delune.TERMCLUSTER)
		self.addClassifier (delune.META)
		
		subclfs = dict ([(cl, obj) for cl, obj in self.clfs.items () if cl in (delune.NAIVEBAYES, delune.FEATUREVOTE, delune.ROCCHIO)])
		self.clfs [delune.MULTIPATH].setopt (clfs = subclfs, cluster = self.clfs [delune.TERMCLUSTER])
		self.clfs [delune.META].setopt (clfs = subclfs)
		
		for classifier, option in self.classifer_opts [:]:
			self.__setopt (classifier, **option)
		
		self.cache.clear ()
		self.closed = 0
		self.mod_time = self.si.getModfiedTime ()			
		return 1

	def do_status (self, *arg, **karg):
		if self.reader:
			segmentinfos = [(self.reader.reader.seg, 0, 0, self.reader.reader.ti.numterm)]		
		else:
			segmentinfos = []			
		locks, note = self.si.lock.locks ()
		subgroup = "%s.ts" % os.path.split (self.si.fs.getmaster ())[-1]
		
		return {
			"lastupdated": self.mod_time,
			"version": self.si.version,			
			"indexdirs": self.si.indexdir,	
			"segmentfiles": {
				"primary": filesystem.get_segment_files ([self.si.fs.getmaster ()]), 
				subgroup: filesystem.get_segment_files ([os.path.join (self.si.fs.getmaster (), subgroup)])
			},
			"numquery": self.numquery,
			"N": self.si.getN (),
			"segmentinfos": segmentinfos,			
			"segmentsizes": self.si.fs.dirinfo (),	
			"locks": locks,
			"note": note			
		}
	
	def guess (self, *arg, **karg):
		default = {"code": 500, "err": "Default Error", "total": 0, "result": []}
		return self.multi_job (self.do_quess, default, *arg, **karg)
	
	def cluster (self, *arg, **karg):
		default = {"code": 500, "err": "Default Error", "total": 0, "result": []}
		return self.multi_job (self.do_cluster, default, *arg, **karg)
		
	def do_delete (self, qs):
		self.reader.delete (qs)		
	
	def do_cluster (self, qs, lang = 'un', *arg, **karg):
		try:
			qs = unistr.makes (qs)
		except:
			return {"code": 501, "err": "Arguments Error", "total": 0, "result": []}		
		if not qs:
			return {"code": 502, "err": "No Keyword", "total": 0, "result": []}		
		if not self.reader:
			return {"code": 503, "err": "No Reader", "total": 0, "result": []}
							
		s = time.time ()
		mem = self.reader.get_memory ()
		clusters = self.get_classifier (delune.TERMCLUSTER).cluster (mem, qs, lang)
		result = {
			"code": 200,
			"time": int ((time.time () - s) * 1000),
			"total": len (clusters),	
			"result": clusters
		}
		return result
			
	RX_SPACE = re.compile ("\s+")
	def do_quess (self, qs, lang = 'un', cl = "naivebayes", top = 0, cond = "", *arg, **karg):
		try:
			top = int (top)
			qs = unistr.makes (qs)
			if cond:
				cond = unistr.makes (cond)			
		except:
			return {"cl": cl, "code": 501, "err": "Arguments Error", "total": 0, "result": []}		
		
		if not qs:
			return {"cl": cl, "code": 502, "err": "No Keyword", "total": 0, "result": []}		
		if not self.reader:
			return {"cl": cl, "code": 503, "err": "No Reader", "total": 0, "result": []}
		
		s = time.time ()
		mem = self.reader.get_memory ()				
		classifier = self.get_classifier (cl)
		if not classifier:
			return {"cl": cl, "code": 504, "err": "No Classifier", "total": 0, "result": []}
			#return [501, 0, 0, 0]
			
		decision = classifier.guess (mem, qs, lang, cond)
		if top == 0:
			top_decision = []
			current_score = None
			for label, score in decision:
				if top_decision and (score < top_decision [-1][-1] or label == top_decision [-1][0]):
					break
				top_decision.append ((label, score))				
			decision = top_decision			
		else:
			decision = decision [0:top]
			
		result = {
			"cl": cl,
			"code": 200,
			"time": int ((time.time () - s) * 1000),
			"total": len (decision),	
			"result": decision
		}		
		return result

		
def main ():
	from delune import indexinfo
	from delune.lib import confparse
	import odbc

	memory.create (1, 32768, 100, "segment")
	inf = indexinfo.IndexInfo (confparse.ConfParse (r"d:\bladepub\proto\etc\col\unspsc.family.class"))
	f = Classifier (inf)
	f.initialize ()

	dbc = odbc.odbc ("pepsi")
	c = dbc.cursor ()
	c.execute ("""
		select top 1300 left (egcc, 4) unspsc, 
		isnull (title, '') + ' ' + isnull (seg, '') + ' ' + isnull (fam, '') + ' '+ isnull (cla, '') + ' ' + isnull (com, '') as description
		from iisd.dbo.trainset_unspsc
		where lang = 0 and substring (egcc, 3, 2) <> '00'		
	""")

	matched = 0
	processed = 0
	unclassified = 0
	agreed = 0
	c1 = 0
	c2 = 0
	c3 = 0

	items = c.fetchall ()
	for item in items [1000:]:	
		processed += 1
		unspsc, title = item
		res = f.query (title, engine = "nb")
		#res = f.guess (title)
		res = res [4:]
		
		if res:
			print(unspsc, res [:2])
			if type (res [0]) == type (()):
				if res [0][0] == unspsc:
					matched += 1
			else:	
				if res [0] == unspsc:
					matched += 1
		else:
			unclassified += 1
	f.close ()
	
	print("_" * 79)
	print((
		"Processed: %d\n"
		"Unclassified: %d\n"
		"Matched: %d\n"
		"Agree: %d\n"
		"1 Class: %d\n"
		"2 Classes: %d\n"
		"3 Classes: %d\n"

		"Accuracy: %2.3f%%\n" % (processed, unclassified, matched, agreed, c1, c2, c3, float (matched) / (processed - unclassified) * 100)
	))

	c.close ()
	dbc.close ()
	

if __name__ == "__main__":
	main ()



import math
import delune
from delune import _delune
from . import naivebayes

class Classifier (naivebayes.Classifier):
	def __init__ (self, reader, analyzer, featset, use_top_features, logger = None):
		self.clfs = {}
		naivebayes.Classifier.__init__ (self, None, None, {}, 0, logger)
				
	def setopt (self, **opts):
		self.clfs = opts.get ("clfs", self.clfs)
		naivebayes.Classifier.setopt (self, **opts)
			
	def match (self, name):
		return name == "meta"
	
	def merge_and_choice (self, res):
		nres = []
		for r in res:
			if not r: continue
			current_score = None
			for label, score in r:	
				if current_score is not None and score < current_score:
					break
				nres.append (label)
				current_score = score
		#print (nres)
		if not nres: return []
		score = 1 / len (nres)
		d = {}		
		for label in nres:
			try: d [label] += score
			except KeyError: d [label] = score

		k = list(d.items ())
		k.sort (key = lambda x: x [1], reverse = True)
		return k			
	
	def guess (self, mem, qs, lang = "un", cond = ""):
		result = []
		for name, classifier in list(self.clfs.items ()):
			try:
				subresult = classifier.guess (mem, qs, lang, cond)
				if not subresult:
					continue					
				
				#print (name, subresult)
				if type (subresult [0]) == type ([]):
					result.extend (subresult)
				else:	
					result.append (subresult)
			except:
				self.logger.trace ("engine")	
		
		return self.merge_and_choice (result)
		


import requests
import queue
import os
import delune
from delune import filesystem
from delune.lib import pathtool, logger as loggerlib, flock
import threading, time
import socket, json
import pickle

mirror_server = None
DUPLICATE_LOCK = "duplicate-%s" % socket.getfqdn ()
DEBUG = True

def normpath (d):
	d = d.replace ("\\", "/")
	if DEBUG:
		d = d.replace ('D:/apps/', 'd:/var/')
	return d

def remove_end_slash (url):
	while url:
		if url [-1] != "/":
			return url
		url = mirror_server [:-1]	
			
def download (session, item):
	global mirror_server
	
	col, entity, group, fn, savepath	= item
	pathtool.mkdir (os.path.split (savepath)[0])
	url = "%s/%s/%s/%s/%s" % (mirror_server, col, entity, group, fn)		
	surl = "/%s/%s/%s/%s" % (col, entity, group, fn)		
	logger ("downloading from %s" % url)
	
	response = session.get(url, stream=True)
	cl = int (response.headers ["Content-Length"])						
	rl = 0
	
	if not response.ok:
		logger ("download failed  HTTP %d: %s" % (response.status_code, url), 'fail')
		return -1
		
	with open(savepath, 'wb') as handle:
		if cl:
			for block in response.iter_content(1024):
				rl += len (block)
				if rl % 1024000 == 0:
					logger ("%d%% downloaded %s" % (rl/cl * 100, surl))
				handle.write(block)				
			logger ("downloaded %d of %d (%d%%) from %s" % (rl, cl, rl/cl * 100, surl))
			
	if rl != cl:
		logger ("incomplete download: %s" % url, 'fail')
		return -1
	return 1			

def mirror ():
	global mirror_server, q, config_dir
	
	remote_session = requests.Session ()
	while 1:
		col = q.get ()
		if not col:
			break
			
		logger ("%s mirroring started" % col)
		remote_session.post ("%s/%s/locks/%s" % (mirror_server, col, DUPLICATE_LOCK))
		r = remote_session.get ("%s/%s" % (mirror_server, col))
		status = r.json ()
		
		local_col = delune.collection (indexdir = [normpath (d) for d in status ['indexdirs']], mode = delune.MODIFY)		
		if local_col.lock.islocked ("replicate"):
			if flock.isCurrentProcess (int (local_col.lock.lockread ("replicate")), 'python'):
				logger ("already under replicate, terminated")
				continue
		local_col.lock.lock ("replicate", str (os.getpid ()))
				
		keep_going = True		
		for name, ctime in status ["locks"]:
			if name == "index":
				logger ("origin %s is under indexing, skip updating index..." % col)
				keep_going = False
				break
		
		if keep_going:
			local_segments = status ["segments"][0]
			if not os.path.isfile (local_segments):
				local_latest_segments = -1				
			else:
				with open (local_segments, "rb") as f:
					temp = pickle.load (f)
				local_latest_segments = max (list(temp.segmentInfo.keys ()))
			origin_segments = [seginfo [0] for seginfo in status ['segmentinfos']]
			
			if not origin_segments:
				logger ("origin %s is empty" % col)
				keep_going = False
			elif max (origin_segments) < local_latest_segments:
				logger ("origin %s is older version, maybe next time..." % col)
				keep_going = False
		
		que = []
		new_file = False		
		if keep_going:
			for d in status ['indexdirs']:
				pathtool.mkdir (normpath (d))
			
			for group in status ["segmentfiles"]:
				segs = status ["segmentfiles"][group]			
				if 'segments' in segs:
					sfn = segs.pop ('segments')[0]
					dfn = normpath (sfn)
					if group == "primary":
						que.append ((col, 'collection', group, "segments", dfn + ".new"))
					else:
						que.insert (0, (col, 'collection', group, "segments", dfn))
				
				for seg, files in segs.items ():
					if group == "primary" and int (seg) not in origin_segments:
						continue
					
					for sfn, ssize, smtime in files:
						dfn = normpath (sfn)
						if not os.path.isfile (dfn):
							que.insert (0, (col, 'segments', group, str(seg) + sfn [-4:], dfn))
							new_file = True
						else:	
							dmtime = os.path.getmtime (dfn)
							dsize = os.path.getsize (dfn)
							if dsize == ssize and smtime <= dmtime:
								continue
							que.insert (0, (col, 'segments', group, str (seg) + sfn [-4:], dfn))
							new_file = True
				
		if new_file:
			failed = 0
			for item in que:
				#col, entity, group, fn, savepath
				if download (remote_session, item) == -1:
					failed += 1
			
			local_segments = status ["segments"][0]
			if failed:
				logger ("%s mirror failed, maybe mext time" % col, "fail")
				os.remove (local_segments + ".new")
			else:
				if os.path.isfile (local_segments):
					os.remove (local_segments)	
				os.rename (local_segments + ".new", local_segments)
				
		remote_session.delete ("%s/%s/locks/%s" % (mirror_server, col, DUPLICATE_LOCK))
		logger ("%s mirroring complete" % col)
		
		logger ("download configuration %s" % col)	
		colopt = status ["colopt"]
		dcolopt = normpath (colopt ['path'])
		if not os.path.isfile (dcolopt) or colopt ["mtime"] > os.path.getmtime (dcolopt) or colopt ["size"] != os.path.getsize (dcolopt):
			with open (dcolopt, "w") as out:
				out.write (json.dumps (colopt ['data']))			
			logger ("%s patch collection configure" % col)
		
		logger ("%s clean up unused segments" % col)		
		local_col.removeDeletables ()
		local_col.lock.unlock ("replicate")
		
		
if __name__ == "__main__":
	import sys, getopt
	
	logger = loggerlib.screen_logger ()
	argopt = getopt.getopt(sys.argv[1:], "l:", ["log="])
	for k, v in argopt [0]:
		if k == "-l" or k == "--log":
			logger = loggerlib.multi_logger ([logger, loggerlib.rotate_logger (v, "delune", "daily")])
	
	if len (argopt [1])	 == 1:
		origin = argopt [1][0]		
	else:	
		print ('Usage: mirror.py origin')
		print ('  ex. mirror.py  http://192.168.1.200:5000/v1')
		sys.exit ()	
	
	mirror_server = remove_end_slash (origin)		
	session = requests.Session ()
	origin_side = session.get ("%s/" % mirror_server).json () ["collections"]
	if not origin_side:
		logger ("no collection to replicate" % col)
		sys.exit ()
	
	r = session.get ("%s/%s" % (mirror_server, origin_side [0]))
	config_dir = os.path.dirname (r.json ()["colopt"]["path"])
	
	delune.configure (1, logger)
	q = queue.Queue ()
	for col in origin_side:		
		q.put (col)
	
	q.put (None)
	t = threading.Thread (target = mirror)
	t.start ()
	
	q.put (None)
	mirror ()
	t.join ()
	
	logger ("syncing collections")	
	for col in os.listdir (config_dir):
		if col not in origin_side:
			os.rename (os.path.join (config_dir, col), os.path.join (config_dir, "-" + col))
			logger ("collection %s has been removed" % col)	
	logger ("all replication process is done.")		
	


#encoding: utf8
import re
from . import typeinfo
from .. import querylexer
import delune
 
class Field:
	methology = "Term"
	def __init__ (self, mem, segment, type, term, lang, name, qstr, num, analyzer = None, loadnorm = False, ngram_opts = ""):
		self.mem = mem
		self.segment = segment
		self.type = type
		self.term = term
		self.lang = lang
		self.name = name
		self.qstr = qstr
		self.num = num
		self.analyzer = analyzer
		self.loadnorm = loadnorm
		self.ngram_opts = ngram_opts
		
		self.value = None
		self.readprox = 0
		self.stopword = 0
		self.near = 0
		self.loose = 0
		self.weight = 1.0
		self.ngram = 0
		self.ti = None
		self.sp = None
		self.df = 0
		
		if self.term: self.process ()
	
	def get_ngram_opts (self, qstr):
		ngram_opts = ""
		if qstr [-1] == "|":
			ngram_opts += "R"
			qstr = qstr [:-1]
		if qstr [0] == "|":
			ngram_opts += "L"
			qstr = qstr [1:]
		return qstr, ngram_opts
		
	def loadTi (self):
		if not self.type: return
		self.ti = self.segment.getTermInfo (self.mem, self.value, self.num)
		if self.ti:
			self.df = self.ti [0]
		
	def loadSm (self):
		self.sp = self.segment.getSortMapPointer (self.num)
	
	def setTi (self, ti, sp = None):
		self.ti = ti
		self.df = ti [0]
		self.sm = sp
					
	def __str__ (self):
		return self.term.decode("utf8").encode ("euckr")
	
	def __repr__ (self):
		return "<%s %s analyzed: %s (fdtype: %s, fdno: %s) [readprox:%d, near:%d, loose:%d]>" % (
			self.methology, 
			self.term.encode ("euckr"), 
			self.qstr.encode ("euckr"),			
			self.type, self.num, self.readprox, self.near, self.loose
		)
	
	def __cmp__ (self, other):
		return self.segment == other.segment and self.num == other.num and self.value == other.value
		
	def process (self):
		self.value = self.qstr
		self.loadTi ()
		
		
class Coord (Field):
	methology = "Coord"
	def process (self):
		self.loadSm ()
		if not self.sp: return
		try:
			coord, meter = self.qstr.split ("~")
			meter = int (meter) # convert from km to m
		except ValueError:
			return
		
		precision = 10 ** (typeinfo.typemap.getsize (self.type)-2)	
		try:
			lat, lon = [int ((float (x) + 180) * precision) for x in coord.split ("/")]
		except ValueError:
			return
		
		self.value = (lat, lon, meter)

class Digit (Field):
	methology = "Digit"
	def process (self):
		self.loadSm ()
		if not self.sp: return
		if self.qstr == "all":
			a = -1
			b = -1
		
		else:
			try: 
				a, b = self.qstr.split ("/")
			except ValueError:
				try:
					a, b = self.qstr.split ("~")
				except ValueError:
					try: 
						a = int (self.qstr)
					except:
						return 0
					else:
						if typeinfo.typemap.isbit (self.type):
							a = str (self.qstr)
							b = "all"
						else:
							a = int (self.qstr)
							b = a + 1
		
		if b in ("all", "any", "none"):
			try: a = int (a, 2)
			except: return
			self.smfunc = 0
			
		else:
			if a == "":
				a = -1
			else:	
				try: a = int (a)
				except: return
			
			if b == "":
				b = -1		
			try: b = int (b)
			except: return
								
			self.smfunc = 1
		
		self.value = (a, b)


class Term (Field):
	methology = "Term"
	def process (self):
		if (self.analyzer and self.qstr != "it") or type (self.qstr) is type ([]):
			# ngram check
			if type (self.qstr) is type ([]):
				value = self.qstr # maybe recieved from WildCard
				ngram_opts = self.ngram_opts
								
			else: # string type
				qstr, ngram_opts = self.get_ngram_opts (self.qstr)
				value = self.analyzer.analyze (qstr, self.lang, "query")
				if not value:
					self.stopword = 1
					return
			
			if len (value) > 1:
				self.ngram = 1
				
			if self.ngram and len (value) > 2: # ngram but 1 char is not applied
				if "R" not in ngram_opts:
					value = value [:-1] # right wildcard search, if has r, right exactly matching serach
				if "L" not in ngram_opts:
					value = value [1:] # left wildcard
			
			# ngram handle
			if self.ngram:
				self.df = 1 # dummy true value
				self.value = []
				for ngram in value: # for right wildcard search					
					termTi = Term (self.mem, self.segment, self.type, self.name+":"+ngram, self.lang, self.name, ngram, self.num, None, self.loadnorm)
					if not termTi.df:
						self.df = 0
						self.value = []
						return
											
					termTi.readprox = 1
					termTi.near = 1
					termTi.loose = 1
					self.value.append (termTi)

				return # value is list, do not handle any more
			
			# not ngram: should be ascii or ONE char ngram
			self.value = value [0] # first stemmed term
				
		else:
			# needn't analyzing, already analyzed.
			self.value = self.qstr
		
		self.loadTi ()
		if self.loadnorm: self.loadSm ()
	
											
class WildCard (Term):
	methology = "WildCard"
	def process (self):
		qstr, ngram_opts = self.get_ngram_opts (self.qstr)
		test_value = self.analyzer.analyze (qstr, self.lang, "query") # check ngram		
		if len (test_value) > 1: # Ngram is already based on wildcard search
			termTi = Term (self.mem, self.segment, self.type, self.name+":"+self.qstr, self.lang, self.name, test_value, self.num, None, self.loadnorm, ngram_opts)
			self.value = termTi.value
			self.methology = "TermList"
			return
		
		self.ti = self.segment.getTermInfos (self.mem, self.qstr, self.num)		
		if not self.ti: return
		
		if self.loadnorm: self.loadSm ()				
		self.value = []
		for ti in self.ti:
			f = Term (self.mem, self.segment, self.type, None, self.lang, self.name, None, self.num, None, 0)
			f.setTi (ti, self.sp)
			self.value.append (f)

		
class FieldWildCard (Term):
	methology = "WildCard"
	def process (self):
		self.value = self.qstr
		self.ti = self.segment.getTermInfos (self.mem, self.value, self.num)
		if not self.ti: 
			return
		
		self.value = []
		for ti in self.ti:
			f = Field (self.mem, self.segment, self.type, None, self.lang, self.name, None, self.num, None, 0)
			f.setTi (ti)
			self.value.append (f)
		
class TermList (Term):
	methology = "TermList"
	def process (self):
		self.value = []
		optimize = True
		terms = querylexer.TermLexer (self.qstr).tokenize ()
		if not terms:
			self.stopword = 1
			return
		
		for term in terms:			
			if term.find ("^") == 0:
				continue
				
			else:
				termTi = Term (self.mem, self.segment, self.type, self.name+":"+term, self.lang, self.name, term, self.num, self.analyzer, self.loadnorm)
				if termTi.stopword:
					continue
				if not termTi.value:
					self.df = 0
					self.value = []
					return
					
				if termTi.ngram:
					optimize = False # ngram is phrase, no sorting
					self.methology = "TermList"
					self.value += termTi.value
				else:
					self.value.append (termTi) # termTi.value is String
		
		if not self.value: 
			self.stopword = 1			
		elif len (self.value) > 1 and optimize:			
			self.value.sort (key = lambda x: x.df)
		
class Phrase (Term):
	methology = "TermList"
	def process (self):		
		self.value = []
		readprox = 0
		if self.type == delune.TEXT:
			readprox = 1
		
		# relexing for analyzer
		terms = querylexer.TermLexer (self.qstr).tokenize ()
		if not terms:
			self.stopword = 1
			return
			
		nextnear = 1
		nextloose = 0		
		for term in terms:
			if not term: continue			
			if term.find ("^") == 0:
				if self.type == delune.TEXT:
					if not self.value: continue
					nextnear = int (term [1:])
					if nextnear > 1: nextloose = 1					
				continue
							
			termTi = Term (self.mem, self.segment, self.type, self.name+":"+term, self.lang, self.name, term, self.num, self.analyzer, self.loadnorm)
			if termTi.stopword: 
				continue
			
			if not termTi.value:
				self.value = []
				self.df = 0
				return
			
			if termTi.ngram:
				termTi.value[0].readprox = readprox
				# 띄어쓰기문제 간격 2으로 탐색 허용
				# to skip '지>' and '<창'
				termTi.value[0].near = 2
				termTi.value[0].loose = 1
				self.value += termTi.value
			else:  # termTi.value is String
				termTi.readprox = readprox
				termTi.near = nextnear
				termTi.loose = nextloose	
				self.value.append (termTi)
			
			nextnear = 1
			nextloose = 0
		
		if not self.value: 
			self.stopword = 1		
		
	
					
class FieldReader:
	def __init__ (self, mem, segment, analyzer, loadnorm):
		self.mem = mem
		self.segment = segment
		self.analyzer = analyzer		
		self.loadnorm = loadnorm
		
	def __call__ (self, type, term, lang, name, qstr, num):
		if not type:
			return Field (self.mem, self.segment, type, term, lang, name, qstr, num)
		
		if typeinfo.typemap.iscoord (type):
			return Coord (self.mem, self.segment, type, term, lang, name, qstr, num)
		
		elif typeinfo.typemap.isint (type) or typeinfo.typemap.isbit (type):
			return Digit (self.mem, self.segment, type, term, lang, name, qstr, num)
		
		elif typeinfo.typemap.hasnorm (type):
			if qstr [0] == "`":
				qstr = qstr [1:]
				return Phrase (self.mem, self.segment, type, term, lang, name, qstr, num, self.analyzer, self.loadnorm)			
			elif qstr.find ("-") != -1 or qstr.find ("/") != -1 or qstr.find ("^") != -1:
				return Phrase (self.mem, self.segment, type, term, lang, name, qstr, num, self.analyzer, self.loadnorm)			
			elif qstr [-1] == "*":
				qstr = qstr [:-1]				
				return WildCard (self.mem, self.segment, type, term, lang, name, qstr, num, self.analyzer, self.loadnorm)
			else:	
				return TermList (self.mem, self.segment, type, term, lang, name, qstr, num, self.analyzer, self.loadnorm)
				
		else:
			if qstr [0] == "`":
				qstr = qstr [1:]			
			if qstr [-1] == "*": # wildcard is not perform analysis
				qstr = qstr [:-1]
				return FieldWildCard (self.mem, self.segment, type, term, lang, name, qstr, num)			
			else:	
				return Field (self.mem, self.segment, type, term, lang, name, qstr, num)
		
		
if __name__ == "__main__":
	from delune.analyzer import porterstemAnalyzer
	
	f = porterstemAnalyzer.Analyzer ()
	x = FieldReader (None, None, f, 1)
	ti = x ("index", 'asdasda', "default", 'asdadads', 0)
	print(ti.value)
	
	
	


import types
from .. import memory
from ..util import util

class Fetcher:
	DEBUG = 0
	def __init__ (self, segments, bucket, summarize, highlights, returns, nthdoc, encoding = None):
		self.segments = segments
		self.bucket = bucket
		self.encoding = encoding
		
		if type (highlights) is bytes:
			self.highlights = highlights.split ("|")
		elif type (highlights) is dict:
			self.highlights = list(highlights.keys ())
		else:
			self.highlights = highlights
		
		self.returns = returns
		self.nthdoc = nthdoc
		self._returns = None
		self.summarize = summarize
		self.fetched_index = 0
		self.fetchmax = len (self.bucket)

	def close (self):
		pass

	def __fetch (self, seg, docid, extra, score):
		mem = memory.get (self.segments [seg]) #get mutex
		field, summary = self.segments [seg].getDocument (mem, docid, self.summarize, self.highlights, self.nthdoc)
		field = util.deserialize (field)
		if self.returns:
			if self._returns is None:
				self._returns = [x.strip () for x in self.returns.split (",")]
				if type (field) is list:
					self._returns = [int (x) for x in self._returns]
			nfield = []
			for i  in self._returns:
				nfield.append (field [i])
			field = nfield

		row = [field]
		if self.summarize:
			if self.encoding:
				summary = summary.encode (self.encoding)
			row.append (summary)
		else:
			row.append ("")
		row.append (seg)
		row.append (docid)
		row.append (extra)
		row.append (score)
		return row

	def fetchmany (self, count):
		result = []
		for hit in self.bucket [self.fetched_index: self.fetched_index + count]:
			field = self.__fetch (*hit)
			result.append (field)

		self.fetched_index += count
		return result

	def fetchone (self):
		if self.fetched_index >= self.self.fetchmax:
			return None

		field = self.__fetch (self.bucket [self.fetched_index] *hit)
		self.fetched_index += 1
		return field

	def fetchall (self):
		return self.fetchmany (self.fetchmax)



import os
import sys
import types
from delune import _delune
from delune.searcher import indexer
from ..util import util
from ..classifier.segment import composedsegmentwriter
import math
import delune
from ..classifier import trainer


# Cosine Similarity of Two Documents
#														  SIGMA ((TFIDF of Da,i) * (TFIDF of Db,i))
# sim (Da, Db) = --------------------------------------------------------------------------
#								SQRT (SIGMA ((TFIDF of Da,i) ** 2)) * SQRT (SIGMA ((TFIDF of Db,i) ** 2))
# 
# where Dai is word Wi in All Word Vector

	
class Corpus:
	N = 0
	pools = []
	numpools = 0
	poolno = {}
	terms = {}
	scores = []
	features = {}

				
class TermCluster (trainer.Trainer):	
	def train (self, classifier = None, **options):
		if classifier is None:
			classifier = "default"
			
		selector = options.get ("selector", None)
		if not selector and classifier == "default":
				raise ValueError("No default selector, be specify")
				
		if selector:
			select_way = options.get ("select_way", delune.SUM)
			select = float (options.get ("select_ratio", 1.0))
			prune_df_min = int (options.get ("prune_df_min", 0))
			prune_df_max = int (options.get ("prune_df_max", 10))
			self.selectFeatures (classifier, select, selector, select_way, prune_df_min, prune_df_max)		
					
	def build (self, min_df = 0):
		if self.indexer is not None:
			self.indexer.commit (1)
			self.indexer.merge ()
			self.indexer.close ()
			self.indexer = None
			
		if not self._initialized:
			self.init ()
		
		self.si.log ("building corpus data")
		self.writer.dirty = 1
		breader = self.writer.getSegmentBulkReader ()
		# get self.corpus pools and DF
		tindex = 0
		pruned = 0
		total = -1
		stat = {}
		while 1:
			try:
				ti = breader.advanceTermInfo ()
				if ti.fdno == self.writer.fdnoterm:
					total += 1
					if min_df and ti.df <= min_df:
						pruned += 1
					else:	
						self.corpus.terms [ti.term] = tindex
						stat [ti.term] = ti.df
						tindex += 1					
			except IndexError:
				break		
		breader.close ()
		
		self.si.log ("%d terms pruned term dimension was shrinked: %d->%d" % (pruned, total, total-pruned))
		temp = list(stat.items ())
		temp.sort (key = lambda x: x[1], reverse = True)
		for term, df in temp [:20]:
			self.si.log ("top DF term: %s %d" % (term, df))
		del stat
		del temp
			
		self.corpus.pools.sort ()
		# caching related with pools
		self.corpus.numpools = len (self.corpus.pools)
		pindex = 0
		poolno = {}
		for pool in self.corpus.pools:
			poolno [pool.name] = pindex
			pindex += 1
		self.corpus.poolno = poolno
		# caching related with pools--
		
		c = 0
		t = len (self.corpus.terms)
		for term in self.corpus.terms:
			if c % 1000 == 0:				
				self.si.log ("build corpus %d/%d" % (c, t))
			tdf, fi = self.writer.featureClassInfo (term, self.corpus)
			i = 0
			for df, tf in fi:
				self.corpus.pools [i].tf += tf
				i += 1			
			c +=1
		
		self.writer.commit (self.corpus)
		
	def scoringFeaures (self, selector, select_way, prune_df_min, prune_df_max):
		terms = self.corpus.terms
		sl = _delune.Selector (len (self.corpus.pools), self.corpus.N, len (terms))
		
		if select_way == delune.SUM:		method = 0
		elif select_way == delune.MAX:	method = 1
		elif select_way == delune.AVG:	method = 2
		
		selector = getattr (sl, selector)		
		self.si.log ("setting corpus")
		fi = []
		for pool in self.corpus.pools:
			fi.append ((pool.df, pool.tf))
		sl.addCorpus (fi)

		self.si.log ("calculating feature's score...")
		scores = []
		t = len (terms)
		c = 0
		p = 0
		for term in terms:
			c += 1
			if c % 1000 == 0:
				self.si.log ("calculating feature %d/%d (%d pruned)" % (c, t, p))
				
			tdf, fi = self.writer.featureClassInfo (term, self.corpus)
			if not tdf: continue
			if prune_df_min > 0 and tdf <= prune_df_min:
				p += 1
				continue
			if prune_df_max > 0 and tdf >= prune_df_max:
				p += 1
				continue			
			sl.add (fi)
			score = selector (method)
			scores.append ((term, score))			
		
		self.si.log ("sorting by score")
		scores.sort (key = lambda x: x[1], reverse = True)
		self.si.log ("feature selection, done.")
		sl.close ()
		return scores
	
	def selectFeatures (self, name, select, selector, select_way, prune_df_min = 0, prune_df_max = 0):
		self.si.log ("selecting [%s] using [%s:%s] for [%s]" % (select, selector, select_way, name))

		scores = self.scoringFeaures (selector, select_way, prune_df_min, prune_df_max)
		if not select:
			select = 1.0
		if select <= 1.0:
			select = int (len (scores) * select)
		else:
			select = int (select)
		if len (scores) < select:
			select = len (scores)
		
		for term, score in scores [:20]:
			self.si.log ("top scored term: %s %f" % (term, score))
				
		self.si.log ("selecting %d features" % (select,))
		terms = scores [:select]
		terms = [x [0] for x in terms]
		
		fmap = dict ([(term, self.corpus.terms [term]) for term in terms])
		if name == "all":
			self.writer.si.segments.features = fmap
		else:	
			self.writer.si.segments.features [name] = fmap
	
	def getFeatureSet (self, name):
		return self.writer.si.segments.features	[name]	
			
	
def usage ():
	print("""
Usage:
	%s [long option list...]
	--config=configuration file path
	--force
	--index
		sub option: --[bulk | append]	
	""" % sys.argv [0])
	sys.exit ()
	
	
def main (path, index = 0, rebuild = 0, force = 0):
	util.check_config (path)
	f = Trainer (path)
	if f.indexable (force):
		f.main (index, rebuild, force)


if __name__ == "__main__":
	import sys, getopt
	import profile

	try: argopt = getopt.getopt(sys.argv[1:], "", \
		["config=", "index", "bulk", "append", "force"])
	except:
		usage ()

	_config = None
	_rebuild = None
	_force = 0
	_index = 0

	for k, v in argopt[0]:
		if k == "--config":
			_config = v
		elif k == "--index":
			_index = 1
		elif k == "--bulk":
			_rebuild = 1
		elif k == "--append":
			_rebuild = 0
		elif k == "--force":
			_force = 1

	if not _config: usage ()
	if _index and _rebuild is None: usage ()

	main (_config, _index, _rebuild, _force)	



from .. import _delune, colsetup
from .segment import typeinfo
import types
from . import searcher, indexer

class Segments (colsetup.Segments):
	def __init__ (self, alias = "", version = None):
		colsetup.Segments.__init__ (self, alias, version)
		self.N = 0
		self.lastIndexedId = 0
		self.fieldNum = -1
		self.fieldInfo = {}
		self.primary = None		
		self.weightMap = {}
		

class Collection (colsetup.CollectionSetup):
	exts = ["tfq", "prx", "del", "tii", "tis", "fda", "fdi", "smp"]
	segment_class = Segments
	
	def __init__ (self, indexdir, mode = 'r', analyzer = None, logger = None, plock = None, ident = None, version = 1):
		colsetup.CollectionSetup.__init__ (self, indexdir, mode, analyzer, logger, plock, ident, version)
		self.cache_type = {}
		self.cache_sortmap = []
	
	def getSortMapFields (self):
		if not self.cache_sortmap:
			for type, num in list(self.segments.fieldInfo.values ()):
				tlen = typeinfo.typemap.getsize (type)
				if tlen:
					self.cache_sortmap.append ((num, tlen)) #fieldnum, size
				elif typeinfo.typemap.hasnorm (type):
					self.cache_sortmap.append ((num, 1))
						
			self.cache_sortmap.sort (key = lambda x: x[0])
				
		return self.cache_sortmap
		
	def getType (self, fdno):
		if not self.cache_type:
			for t, n in list(self.segments.fieldInfo.values ()):
				self.cache_type [n] = t
				
		return self.cache_type [fdno]
	
	def getTypeByName (self, fieldName):
		try: 
			return self.segments.fieldInfo [fieldName][0]
		except KeyError:
			return
	
	def setWeight (self, field, weight = 1.0):
		if type (field) is dict:
			for _name, _weight in list(field.items ()):
				self.segments.weightMap [_name] = _weight				
		else:
			self.segments.weightMap [field] = weight		
		
	def getWeight (self, fieldName):
		try:
			return float (self.segments.weightMap [fieldName])
		except KeyError:
			return 1.0		
	
	def getFieldInfo (self, fieldName):
		try:
			fieldNum = self.segments.fieldInfo [fieldName][-1]
		except KeyError:
			return "", ""
			
		fieldType = self.getType (fieldNum)		
		return fieldType, fieldNum
		
	def getFdnoByName (self, fieldName):		
		try:
			return self.segments.fieldInfo [fieldName][1]
		except KeyError:
			return
	
	def setPrimary (self, fdname):
		self.segments.primary = fdname
	
	def getPrimary (self):
		return self.segments.primary
		
	def setLastIndexedId (self, id):
		self.segments.lastIndexedId = id
		
	def getLastIndexedId (self):
		return self.segments.lastIndexedId
		
	def setFieldInfo (self, fieldName, type):
		try: 
			return self.segments.fieldInfo [fieldName][1]
		except KeyError: 
			self.segments.fieldNum += 1
			self.segments.fieldInfo [fieldName] = (type, self.segments.fieldNum)
			self.cache_type = {} # expire
			self.cache_sortmap = [] # expire
			return self.segments.fieldNum
	
	def getN (self):
		return self.segments.N
	
	def get_indexer (self, **karg):
		self.setopt (**karg)
		return indexer.Indexer (self)
	
	def get_searcher (self, **karg):
		self.setopt (**karg)
		return searcher.Searcher (self)
	

		
		